from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import (
    Column,
    Integer,
    ForeignKey,
    String,
    LargeBinary,
    Boolean,
    UniqueConstraint,
    Index,
    Enum,
    select,
    TIMESTAMP,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship

from qcfractal.db_socket import BaseORM
from qcportal.auth import AuthTypeEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from typing import Optional, Iterable, Dict, Any


class UserGroupORM(BaseORM):
    """
    Table for storing which groups a user belongs to
    """

    __tablename__ = "user_groups"

    user_id = Column(Integer, ForeignKey("user.id", ondelete="cascade"), nullable=False, primary_key=True)
    group_id = Column(Integer, ForeignKey("group.id", ondelete="cascade"), nullable=False, primary_key=True)


class GroupORM(BaseORM):
    """
    Table for storing group information
    """

    __tablename__ = "group"

    id = Column(Integer, primary_key=True)
    groupname = Column(String, nullable=False)
    description = Column(String, nullable=False)

    __table_args__ = (UniqueConstraint("groupname", name="ux_group_groupname"),)


class UserORM(BaseORM):
    """
    Table for storing user information
    """

    __tablename__ = "user"

    id = Column(Integer, primary_key=True)

    role = Column(String, nullable=False)

    groups_orm = relationship(GroupORM, secondary=UserGroupORM.__tablename__)

    auth_type = Column(Enum(AuthTypeEnum), nullable=False)

    username = Column(String, nullable=False)
    password = Column(LargeBinary, nullable=False)
    enabled = Column(Boolean, nullable=False, server_default="true")
    fullname = Column(String, nullable=False, server_default="")
    organization = Column(String, nullable=False, server_default="")
    email = Column(String, nullable=False, server_default="")

    __table_args__ = (UniqueConstraint("username", name="ux_user_username"),)

    _qcportal_model_excludes = ["groups_orm", "password"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)

        d["groups"] = [x.groupname for x in self.groups_orm]
        return d


class UserSessionORM(BaseORM):
    """
    Table for storing flask session information
    """

    __tablename__ = "user_session"

    public_id = Column(Integer, primary_key=True)
    session_key = Column(String, nullable=False)
    user_id = Column(Integer, ForeignKey("user.id", ondelete="cascade"), nullable=False)
    session_data = Column(JSONB, nullable=False)
    last_accessed = Column(TIMESTAMP(timezone=True), nullable=False, default=now_at_utc)

    __table_args__ = (
        UniqueConstraint("session_key", name="ux_user_session_session_key"),
        Index("ix_user_session_user_id", "user_id"),
    )

    def public_dict(self) -> Dict[str, Any]:
        return {
            "public_id": self.public_id,
            "user_id": self.user_id,
            "last_accessed": self.last_accessed,
            "ip_address": self.session_data.get("ip_address", None),
            "user_agent": self.session_data.get("user_agent", None),
        }


class UserPreferencesORM(BaseORM):
    """
    Table for storing user preference information
    """

    __tablename__ = "user_preferences"

    user_id = Column(Integer, ForeignKey("user.id", ondelete="cascade"), primary_key=True)
    preferences = Column(JSONB, nullable=False)


_user_id_map_subq = select(UserORM.id.label("id"), UserORM.username.label("username")).subquery()


class UserIDMapSubquery(BaseORM):
    __table__ = _user_id_map_subq



from __future__ import annotations

import logging
import os
import pathlib
import re
import secrets
import shutil
import subprocess
import time
from typing import TYPE_CHECKING

import psycopg2
import tabulate
from psycopg2.errors import OperationalError, ObjectInUse

from .config import DatabaseConfig
from .db_socket.socket import SQLAlchemySocket
from .port_util import find_open_port, is_port_inuse

if TYPE_CHECKING:
    from typing import Any, List, Optional, Tuple, Dict
    import psycopg2.extensions


class PostgresHarness:
    def __init__(self, config: DatabaseConfig):
        """A manager for postgres server instances

        This class is used to create, start and stop postgres instances, particularly if
        QCFractal is expected to own this instance. This can also be used to detect
        and update postgres databases.

        Parameters
        ----------
        config
            The configuration options
        """
        self.config = config
        self._logger = logging.getLogger("PostgresHarness")
        self._alembic_ini = os.path.join(os.path.abspath(os.path.dirname(__file__)), "alembic.ini")

        # Figure out the directory containing postgres utilities
        # if not specified in the configuration
        # This is only required if own == True
        self._tool_dir = config.pg_tool_dir if config.own is True else None

        # Did we start the database (and therefore should shut it down?)
        self._started_db = False

        # Own the database, but no tool directory specified
        if config.own and config.pg_tool_dir is None:
            # Find pg_config. However, we should try all the possible pg_config available in the PATH.
            # We depend on libpq, which will install pg_config but none of the other tools
            env_path = os.environ.get("PATH", None)

            # Not sure this should ever happen
            search_dirs = list() if env_path is None else env_path.split(os.pathsep)

            # Go through all the path directories, finding pg_config
            # Then, after running pg_config --bindir, test that path for pg_ctl
            for search_dir in search_dirs:
                self._logger.debug(f"Searching for pg_config in path {search_dir}")
                pg_config_path = shutil.which("pg_config", path=search_dir)
                if pg_config_path is None:
                    continue

                self._logger.debug(f"Found pg_config in path {search_dir}")

                # Run pg_config to get the path
                ret, stdout, _ = self._run_subprocess([pg_config_path, "--bindir"])
                if ret != 0:
                    self._logger.error(f"pg_config returned non-zero error code: {ret}")
                else:
                    possible_path = stdout.strip()

                    # Does pg_ctl exist there?
                    pg_ctl_path = shutil.which("pg_ctl", path=possible_path)
                    if pg_ctl_path is not None:
                        self._logger.info(f"Using Postgres tools found via pg_config located in {possible_path}")
                        self._tool_dir = possible_path
                        break

            # If the tool dir is still None, that is a problem (when own = True)
            if self._tool_dir is None:
                raise RuntimeError(
                    "Postgresql tools cannot be found. Postgresql may not be installed, or pg_tool_dir is incorrect. If you do have Postgresql, try setting pg_tool_dir in the configuration to the directory containing psql, pg_ctl, etc"
                )

            # Determine version
            pg_ctl_path = self._get_tool("pg_ctl")
            ret, stdout, stderr = self._run_subprocess([pg_ctl_path, "--version"])
            if ret != 0:
                raise RuntimeError(
                    f"Error running pg_ctl --version. Something is probably pretty wrong. Return code {ret}\noutput:\n{stdout}\nstderr:\n{stderr}"
                )

            pg_version = stdout.strip()
            self._logger.info(f"Postgresql version found: {pg_version}")

    def _get_tool(self, tool: str) -> str:
        """
        Obtains the path to a postgres tool (such as pg_ctl)

        If 'pg_tool_dir' is not specified, the current PATH environment is searched instead

        If the tool is not found, an exception is raised

        Parameters
        ----------
        tool
            Tool to search for (ie, psql)

        Returns
        -------
        :
            Full path to the tools executable
        """

        tool_path = shutil.which(tool, path=self._tool_dir)
        if tool_path is None:
            raise RuntimeError(
                f"Postgresql tool/command {tool} cannot be found in path {self._tool_dir}. Postgresql may not be installed, or pg_tool_dir is incorrect."
            )
        return tool_path

    def _run_subprocess(
        self, command: List[str], env: Optional[Dict[str, Any]] = None, shell: bool = False
    ) -> Tuple[int, str, str]:
        """
        Runs a command using subprocess, and output stdout into the logger

        Parameters
        ----------
        command
            Command to run as a list of strings (see documentation for subprocess)

        Returns
        -------
        :
            Return code, stdout, and stderr as a Tuple
        """

        full_env = os.environ.copy()

        if env is not None:
            full_env.update(env)

        self._logger.debug("Running subprocess: " + str(command))
        if shell:
            proc = subprocess.run(
                " ".join(command),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=full_env,
                shell=True,
                executable="/bin/bash",
            )
        else:
            proc = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=full_env)

        stdout = proc.stdout.decode()
        stderr = proc.stderr.decode()
        if len(stdout) > 0:
            self._logger.info(stdout)
        if len(stderr) > 0:
            self._logger.info(stderr)

        return proc.returncode, stdout, stderr

    @property
    def database_dsn(self) -> str:
        """Provides the full PostgreSQL connection string (for direct use with psycopg2)"""
        return self.config.psycopg2_dsn

    @property
    def maintenance_dsn(self) -> str:
        """Provides the full PostgreSQL URI string for the maintenance db (for direct use with psycopg2)"""
        return self.config.psycopg2_maintenance_dsn

    def can_connect(self) -> bool:
        """Checks if the postgres instance is alive, that the database exists, and that we can connect to it

        This function swallows most exceptions (OperationalError), and returns False in those cases

        Returns
        -------
        :
            True if the instance is alive and that we can connect to it. Optionally checks that
            the database exists
        """

        try:
            conn = psycopg2.connect(self.database_dsn)
            conn.close()
            return True
        except psycopg2.OperationalError as e:
            self._logger.debug(f"Could not connect to postgres instance: {str(e)}")
            return False

    def is_alive(self) -> bool:
        """Checks if the postgres is alive

        This only checks to see if an instance is running on the given host/port. It does not check
        if we have permission to connect to it or if the database exists

        Returns
        -------
        :
            True if a postgres instance is alive
        """

        try:
            conn = psycopg2.connect(self.database_dsn)
            conn.close()
            return True
        except psycopg2.OperationalError as e:
            estr = str(e)

            if "Connection refused" in estr:
                return False

            # right?
            return True

    def ensure_alive(self):
        """
        Checks to see that the postgres instance is up and running

        If it is not running, but we are expected to control the database, it will be started.
        This function does not check anything about the database.

        This will raise an exception if the database is not alive and/or could not be started
        """

        # if own = True, we are responsible for starting the db instance
        # But don't start it if it is already alive. It may have been started
        # elsewhere.
        if not self.is_alive():
            if self.config.own:
                self.start()
                self._logger.info(f"Started a postgres instance for uri {self.config.safe_uri}")
            else:
                raise RuntimeError(
                    f"A running postgres instance serving uri {self.config.safe_uri} does not appear to be running. "
                    f"It must be running for me to continue "
                )

        self._logger.info(f"Postgres instance serving uri {self.config.safe_uri} appears to be up and running")

    def sql_command(
        self, statement: str, use_maintenance_db: bool = False, autocommit: bool = True, returns=True
    ) -> Any:
        """Runs a single SQL query or statement string and returns the output

        Parameters
        ----------
        statement
            A psql command/query string.
        use_maintenance_db: bool
            If true, connect to the maintenance db instead
        autocommit
            If true, enable autocommit on the connection
        returns
            If true, fetch the results and return them. Set to False for commands that
            don't return anything.
        """

        if use_maintenance_db:
            uri = self.maintenance_dsn
        else:
            uri = self.database_dsn

        conn = psycopg2.connect(uri)
        if autocommit:
            conn.autocommit = True
        cursor = conn.cursor()

        try:
            self._logger.debug(f"Executing SQL: {statement}")
            cursor.execute(statement)
            if returns:
                return cursor.fetchall()
        finally:
            cursor.close()
            conn.close()

    def pg_ctl(self, cmds: List[str]) -> Tuple[int, str, str]:
        """Runs a pg_ctl command and returns its output

        Parameters
        ----------
        cmds
            A list of PostgreSQL pg_ctl commands to run.

        Returns
        -------
        :
            The return code, stdout, and stderr returned by pg_ctl
        """

        # pg_ctl should only be run if we are managing the db
        assert self.config.own
        self._logger.debug(f"Running pg_ctl command: {cmds}")

        pg_ctl = self._get_tool("pg_ctl")
        all_cmds = [pg_ctl, "-l", self.config.logfile, "-D", self.config.data_directory]
        all_cmds.extend(cmds)
        return self._run_subprocess(all_cmds)

    def get_postgres_version(self) -> str:
        """Returns the version of the postgres instance"""

        return self.sql_command("SELECT version()", use_maintenance_db=True)[0][0]

    def get_alembic_version(self) -> str:
        """Returns the version of the alembic schema in the database"""

        return self.sql_command("SELECT version_num from alembic_version")[0][0]

    def create_database(self, create_tables: bool):
        """Creates a new qcfractal database (and tables)

        The postgres instance must be initialized and running.

        If the database is existing, no changes to the database are made.
        If there is an error, an exception in raised

        Parameters
        ----------
        create_tables
            If true, create the tables (schema) as well
        """

        # First we connect to an existing database
        # The database specified in the config may not exist yet
        conn = psycopg2.connect(self.maintenance_dsn)
        conn.autocommit = True

        cursor = conn.cursor()

        try:
            # Now we can search the pg_catalog to see if the database we want exists yet
            cursor.execute(f"SELECT 1 FROM pg_catalog.pg_database WHERE datname = '{self.config.database_name}'")
            exists = cursor.fetchone()

            if not exists:
                self._logger.info(f"Creating database {self.config.database_name}...")
                cursor.execute(f"CREATE DATABASE {self.config.database_name}")
                self._logger.info(f"Database {self.config.database_name} created")

                if create_tables:
                    SQLAlchemySocket.create_database_tables(self.config)
            else:
                self._logger.info(f"Database {self.config.database_name} already exists, so I am leaving it alone")

            # Check to see that everything is ok
            if not self.can_connect():
                raise RuntimeError("I created the database, but now can't connect to it? Maybe check the postgres logs")
        finally:
            cursor.close()
            conn.close()

    def delete_database(self) -> None:
        """
        Deletes the database

        This will delete all data associated with the database!
        """
        # First we connect to the maintenance database
        # The database specified in the config may not exist yet
        conn = psycopg2.connect(self.maintenance_dsn)
        conn.autocommit = True

        cursor = conn.cursor()

        # Now we can search the pg_catalog to see if the database we want exists yet
        self._logger.info(f"Deleting/Dropping database {self.config.database_name}")

        try:
            cursor.execute(f"DROP DATABASE IF EXISTS {self.config.database_name}")
        except (OperationalError, ObjectInUse) as e:
            err = f"Could not delete database. Was it still open somewhere?\nError: {str(e)}\n"
            cursor.execute("SELECT pid,state,query_start,wait_event_type,wait_event,query FROM pg_stat_activity")

            err += "Open Connections\n-----------------------\n"
            err += tabulate.tabulate(
                cursor, headers=["pid", "state", "query_start", "wait_event_type", "wait_event", "query"]
            )
            err += "\n"
            raise RuntimeError(err)
        finally:
            cursor.close()
            conn.close()

    def start(self) -> None:
        """
        Starts a PostgreSQL server based off the current configuration parameters.

        The PostgreSQL server must be initialized and the configured port available.
        The database does not need to exist.
        """

        # We should only do this if we are in charge of the database itself
        assert self.config.own

        # Startup the server
        self._logger.info("Starting the PostgreSQL instance")

        # We should be in charge of this postgres process. If something is running, then that is a problem
        if is_port_inuse("localhost", self.config.port):
            raise RuntimeError(
                f"A process is already running on port {self.config.port} that is not associated with this QCFractal instance's database"
            )
        else:
            retcode, stdout, stderr = self.pg_ctl(["start"])

            err_msg = f"Error starting PostgreSQL. Did you remember to initialize it (qcfractal-server init)?\noutput:\n{stdout}\nstderr:\n{stderr}"

            if retcode != 0:
                raise RuntimeError(err_msg)
            if not (("server started" in stdout) or ("server starting" in stdout)):
                raise RuntimeError(err_msg)

            # Check that we are alive
            for x in range(10):
                if self.is_alive():
                    break
                else:
                    time.sleep(0.2)
            else:
                raise RuntimeError(err_msg)

            retcode, stdout, stderr = self.pg_ctl(["status"])
            if retcode == 0:
                self._logger.info("PostgreSQL successfully started in a background process")
            else:
                err_msg = f"Database seemed to start, but status check failed:\noutput:\n{stdout}\nstderr:\n{stderr}"
                raise RuntimeError(err_msg)

        self._started_db = True

    def shutdown(self) -> None:
        """Shuts down the current postgres instance (if we control it)

        This only shuts down the instance if qcfractal is expected to manage it (own=True) and
        if this process started it
        """

        # We don't manage the database
        if self.config.own is False or self._started_db is False:
            return

        if self.is_alive() is False:
            return

        retcode, stdout, stderr = self.pg_ctl(["stop"])
        if retcode == 0:
            self._logger.info("PostgreSQL successfully stopped")
        else:
            err_msg = f"Error stopping the postgres instance:\noutput:\n{stdout}\nstderr:\n{stderr}"
            raise RuntimeError(err_msg)

    def postgres_initialized(self):
        """
        Returns True if the postgres instance has been initialized, False otherwise
        """

        psql_conf_file = os.path.join(self.config.data_directory, "postgresql.conf")
        return os.path.exists(psql_conf_file)

    def initialize_postgres(self) -> None:
        """Initializes a postgresql instance and starts it

        The data directory and port from the configuration is used for the postgres instance

        This does not create the QCFractal database or its tables
        """

        # Can only initialize if we are expected to manage it
        assert self.config.own

        if not self.config.username or not self.config.password:
            raise RuntimeError("Username or password are not given")

        self._logger.info("Initializing the Postgresql database")

        # Is the specified port open? Stop early if in use
        if is_port_inuse("localhost", self.config.port):
            raise RuntimeError("Port is already in use. Specify another port for the database")

        psql_conf_file = os.path.join(self.config.data_directory, "postgresql.conf")

        if os.path.exists(psql_conf_file):
            raise RuntimeError(f"A config already exists at {psql_conf_file}. Database has been initialized already?")

        initdb_path = self._get_tool("initdb")

        cmd = [initdb_path, "-D", self.config.data_directory, "--auth", "scram-sha-256"]
        if self.config.username is not None:
            cmd += ["--username", self.config.username]

        # Initdb requires passwords come from a file
        env = {"PG_SUPER_PASSWORD": self.config.password}
        cmd += ['--pwfile=<(printf "%s\n" ${PG_SUPER_PASSWORD})']

        retcode, stdout, stderr = self._run_subprocess(cmd, env=env, shell=True)

        if retcode != 0 or "Success." not in stdout:
            err_msg = f"Error initializing a PostgreSQL instance:\noutput:\n{stdout}\nstderr:\n{stderr}"
            raise RuntimeError(err_msg)

        # Change some configuration options
        psql_conf_path = pathlib.Path(psql_conf_file)
        psql_conf = psql_conf_path.read_text()

        assert "#port = 5432" in psql_conf
        psql_conf = psql_conf.replace("#port = 5432", f"port = {self.config.port}")

        # Change the location of the socket file
        # Some OSs/Linux distributions will use a directory not writeable by a normal user
        sock_dir = os.path.join(self.config.data_directory, "sock")
        os.makedirs(sock_dir, exist_ok=True)

        # Only use sockets if the sock_dir path would be less than 103 bytes
        # More is put after the directory, so leave some margin there
        if len(sock_dir) < 80:
            psql_conf = re.sub(
                r"#?unix_socket_directories =.*",
                f"unix_socket_directories = '{sock_dir}'",
                psql_conf,
                flags=re.MULTILINE,
            )
        else:
            psql_conf = re.sub(
                r"#?unix_socket_directories =.*",
                f"unix_socket_directories = ''",
                psql_conf,
                flags=re.MULTILINE,
            )

        psql_conf_path.write_text(psql_conf)

        # Start the postgres instance
        self.start()
        self._logger.info("Postgresql instance successfully initialized and started")

    def backup_database(self, filepath: str) -> None:
        """
        Backs up the database into a file

        Parameters
        ----------
        filepath
            File to store the backup to. Must not exist
        """

        filepath = os.path.realpath(filepath)
        if os.path.exists(filepath):
            raise RuntimeError(f"Path {filepath} exists already, so cannot back up to there")

        cmds = [
            self._get_tool("pg_dump"),
            "-Fc",  # Custom postgres format, fast
            "--dbname",
            self.database_dsn,  # Yes, dbname can take the psycopg2 uri like "host=localhost, port=5432"
            "--file",
            filepath,
        ]

        self._logger.debug(f"pg_backup command: {'  '.join(cmds)}")
        retcode, stdout, stderr = self._run_subprocess(cmds)

        if retcode != 0:
            err_msg = f"Error backing up the database\noutput:\n{stdout}\nstderr:\n{stderr}"
            raise RuntimeError(err_msg)

    def restore_database(self, filepath) -> None:
        if not os.path.exists(filepath):
            raise RuntimeError(f"Backup file {filepath} does not exist or is not a file")

        # Create the database, but not the tables. These are in the backup file, and may represent
        # a different version
        self.create_database(create_tables=False)

        cmds = [
            self._get_tool("pg_restore"),
            "-e",
            "-x",
            "-O",
            "--dbname",
            self.database_dsn,  # Yes, dbname can take the psycopg2 uri like "host=localhost, port=5432"
            filepath,
        ]

        self._logger.debug(f"pg_restore command: {'  '.join(cmds)}")
        retcode, stdout, stderr = self._run_subprocess(cmds)

        if retcode != 0:
            err_msg = f"Error restoring the database\noutput:\n{stdout}\nstderr:\n{stderr}"
            raise RuntimeError(err_msg)

    def database_size(self) -> int:
        """
        Returns the size of the database in bytes
        """

        # sql_command returns a list of tuples
        return self.sql_command(f"SELECT pg_database_size('{self.config.database_name}');")[0][0]


def create_snowflake_postgres(host: str, data_dir: str) -> PostgresHarness:
    """Create and Initialize a postgres instance in a particular directory

    Parameters
    ----------
    host
        The host name or IP address to bind to
    data_dir
        Path to the directory to store the database data
    """

    sock_dir = os.path.join(data_dir, "sock")

    # There is a limit of 103 bytes for a path to a socket file
    # There's extra put after sock_dir, so be conservative and if it's too long,
    # use the regular host
    if len(sock_dir) < 80:
        db_host = sock_dir
    else:
        db_host = host

    port = find_open_port(host)
    db_config = {
        "port": port,
        "data_directory": data_dir,
        "base_folder": data_dir,
        "own": True,
        "host": db_host,
        "username": "qcfractal_snowflake",
        "password": secrets.token_urlsafe(32),
    }

    config = DatabaseConfig(**db_config)
    pg_harness = PostgresHarness(config)
    pg_harness.initialize_postgres()
    return pg_harness



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import (
    Column,
    Integer,
    ForeignKey,
    String,
    JSON,
    Index,
    CheckConstraint,
    UniqueConstraint,
    event,
    DDL,
    select,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import relationship

from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.singlepoint.record_db_models import QCSpecificationORM, SinglepointRecordORM
from qcfractal.db_socket import BaseORM

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Iterable


class OptimizationTrajectoryORM(BaseORM):
    """
    Table for storing optimization to singlepoint relationships (trajectory)
    """

    __tablename__ = "optimization_trajectory"

    optimization_id = Column(Integer, ForeignKey("optimization_record.id", ondelete="cascade"), primary_key=True)
    singlepoint_id = Column(Integer, ForeignKey(SinglepointRecordORM.id), primary_key=True)
    position = Column(Integer, primary_key=True)

    singlepoint_record = relationship(SinglepointRecordORM)

    __table_args__ = (Index("ix_optimization_trajectory_singlepoint_id", "singlepoint_id"),)

    _qcportal_model_excludes = ["optimization_id", "position"]


class OptimizationSpecificationORM(BaseORM):
    """
    Table for storing optimization specifications
    """

    __tablename__ = "optimization_specification"

    id = Column(Integer, primary_key=True)
    specification_hash = Column(String, nullable=False)

    program = Column(String(100), nullable=False)

    qc_specification_id = Column(Integer, ForeignKey(QCSpecificationORM.id), nullable=False)
    qc_specification = relationship(QCSpecificationORM, lazy="joined")

    keywords = Column(JSONB, nullable=False)
    protocols = Column(JSONB, nullable=False)

    __table_args__ = (
        UniqueConstraint("specification_hash", "qc_specification_id", name="ux_optimization_specification_keys"),
        Index("ix_optimization_specification_program", "program"),
        Index("ix_optimization_specification_qc_specification_id", "qc_specification_id"),
        # Enforce lowercase on some fields
        # This does not actually change the text to lowercase, but will fail to insert anything not lowercase
        # WARNING - these are not autodetected by alembic
        CheckConstraint("program = LOWER(program)", name="ck_optimization_specification_program_lower"),
    )

    _qcportal_model_excludes = ["id", "specification_hash", "qc_specification_id"]

    @property
    def required_programs(self) -> Dict[str, Optional[str]]:
        r = {self.program: None}
        r.update(self.qc_specification.required_programs)
        return r

    @property
    def short_description(self) -> str:
        return f"{self.program}+{self.qc_specification.short_description}"


class OptimizationRecordORM(BaseRecordORM):
    """
    Table for storing optimization calculations
    """

    __tablename__ = "optimization_record"

    id = Column(Integer, ForeignKey(BaseRecordORM.id, ondelete="cascade"), primary_key=True)

    specification_id = Column(Integer, ForeignKey(OptimizationSpecificationORM.id), nullable=False)
    specification = relationship(OptimizationSpecificationORM, lazy="selectin")

    initial_molecule_id = Column(Integer, ForeignKey(MoleculeORM.id), nullable=False)
    initial_molecule = relationship(MoleculeORM, foreign_keys=initial_molecule_id)

    final_molecule_id = Column(Integer, ForeignKey(MoleculeORM.id), nullable=True)
    final_molecule = relationship(MoleculeORM, foreign_keys=final_molecule_id)

    energies = Column(JSON)

    trajectory = relationship(
        OptimizationTrajectoryORM,
        order_by=OptimizationTrajectoryORM.position,
        collection_class=ordering_list("position"),
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    __mapper_args__ = {"polymorphic_identity": "optimization"}

    __table_args__ = (
        Index("ix_optimization_record_specification_id", "specification_id"),
        Index("ix_optimization_record_initial_molecule_id", "initial_molecule_id"),
        Index("ix_optimization_record_final_molecule_id", "final_molecule_id"),
    )

    _qcportal_model_excludes = [*BaseRecordORM._qcportal_model_excludes, "specification_id"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseRecordORM.model_dict(self, exclude)

        # Models use ids for the trajectory, rather than the object we store
        if "trajectory" in d:
            d["trajectory_ids"] = [x["singlepoint_id"] for x in d.pop("trajectory")]

        return d

    @property
    def required_programs(self) -> Dict[str, Optional[str]]:
        programs = self.specification.required_programs
        programs["qcengine"] = None
        return programs

    @property
    def short_description(self) -> str:
        return f'{self.initial_molecule.identifiers["molecular_formula"]} {self.specification.short_description}'


# Delete base record if this record is deleted
_del_baserecord_trigger = DDL(
    """
    CREATE TRIGGER qca_optimization_record_delete_base_tr
    AFTER DELETE ON optimization_record
    FOR EACH ROW EXECUTE PROCEDURE qca_base_record_delete();
    """
)

event.listen(
    OptimizationRecordORM.__table__, "after_create", _del_baserecord_trigger.execute_if(dialect=("postgresql"))
)


record_direct_children_select = [
    select(
        BaseRecordORM.id.label("parent_id"),
        OptimizationTrajectoryORM.singlepoint_id.label("child_id"),
    )
    .join(OptimizationTrajectoryORM, BaseRecordORM.id == OptimizationTrajectoryORM.optimization_id)
    .where(BaseRecordORM.record_type == "optimization")
]

record_children_select = record_direct_children_select



import shutil
import tempfile
from functools import wraps
from typing import Callable, Optional, Iterable

from flask import current_app

try:
    import pydantic.v1 as pydantic
except ImportError:
    import pydantic
from flask import request, g, Response
from werkzeug.exceptions import BadRequest

from qcfractal.flask_app import storage_socket
from qcportal.serialization import deserialize, serialize


def get_file_extension(filename, allowed_extensions):
    filename = filename.lower()
    with_period = [f".{ext}" for ext in allowed_extensions]

    for ext in with_period:
        if filename.endswith(ext):
            return ext

    raise BadRequest(f"Invalid file extension on file: {filename}. Allowed extensions: {allowed_extensions}")


def wrap_global_route(
    requested_resource,
    requested_action,
    require_security: bool = False,
    allowed_file_extensions: Optional[Iterable[str]] = None,
) -> Callable:
    """
    Decorator that wraps a Flask route function, providing useful functionality

    This wrapper handles several things:

        1. Checks the JWT for permission to access this route (with the requested action)
           OR
           Checks the session cookie (for browser-based authentication)
        2. Parses the request body and URL params, and converts them to the appropriate model (see below)
        3. Serializes the response returned from the wrapped function into the appropriate
           type (taken from the accepted mimetypes)

    The data packaged with the request may be json, msgpack, or maybe others in the future.
    This is deserialized and converted to the types needed by the wrapped function. These
    types are read from the type annotations on the wrapped function.

    There are two function parameters that are inspected - `url_params` for the URL parameters,
    and `body_data` for data included in the request body. The type annotations for these
    parameters are read, and then pydantic is used to convert the deserialized request body/params
    into the appropriate type, after which they are passed to the function.

    If `allowed_file_extensions` is not None, then this route will also accept file uploads. The wrapped
    function will be passed the file objects as the `files` parameter. If this parameter does not exist,
    an exception will be raised.

    Parameters
    ----------
    requested_resource
        The name of the major resource
    requested_action
        Type of action that this route handles (read, write, etc)
    allowed_file_extensions
        If not None, files are allowed to be uploaded with this route. The files must have these extensions.
    require_security
        If true, route is only accessible if security is enabled on the server.
    """

    def decorate(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            storage_socket.auth.assert_global_permission(g.role, requested_resource, requested_action, require_security)

            ##################################################################
            # If we got here, then the user is allowed access to this endpoint
            # Continue with parsing their request
            ##################################################################

            content_type = request.headers.get("Content-Type")

            # Find an appropriate return type (from the "Accept" header)
            # Flask helpfully parses this for us
            # By default, use plain json
            possible_types = ["text/html", "application/msgpack", "application/json"]
            accept_type = request.accept_mimetypes.best_match(possible_types, "application/json")

            # If text/html is first, then this is probably a browser. Send json, as most browsers
            # will accept that
            if accept_type == "text/html":
                accept_type = "application/json"

            # get the type annotations for body_model and url_params_model
            # from the wrapped function
            annotations = fn.__annotations__

            body_model = annotations.get("body_data", None)
            url_params_model = annotations.get("url_params", None)

            # If files are allowed, check for the files parameter in the function
            if allowed_file_extensions is not None:
                if "files" not in annotations:
                    raise BadRequest('Route allows uploads, but function doesn\'t have the "files" argument')

            # 1. The body is stored in request.data
            #    However, if it's not there, and we have files, it will be in request.files
            #    (since you cannot send body data and files in the same request)
            if "body_data" in request.files:
                body_data = request.files["body_data"].read()
                content_type = request.files["body_data"].content_type
            else:
                body_data = request.data

            if body_model is not None:
                if content_type is None:
                    raise BadRequest("No Content-Type specified")

                if not body_data:
                    raise BadRequest("Expected body, but it is empty")

                try:
                    deserialized_data = deserialize(body_data, content_type)
                    kwargs["body_data"] = pydantic.parse_obj_as(body_model, deserialized_data)
                except Exception as e:
                    raise BadRequest("Invalid body: " + str(e))

            # 2. Query parameters are in request.args
            if url_params_model is not None:
                try:
                    kwargs["url_params"] = url_params_model(**request.args.to_dict(False))
                except Exception as e:
                    raise BadRequest("Invalid request arguments: " + str(e))

            # 3. File uploads
            temp_dir = None
            if allowed_file_extensions is not None:
                kwargs["files"] = []

                if "files" in request.files:
                    temp_dir = tempfile.mkdtemp(dir=current_app.config["UPLOAD_FOLDER"])

                    req_files = request.files.getlist("files")
                    for f in req_files:

                        # Will raise on invalid file extension
                        ext = get_file_extension(f.filename, allowed_file_extensions)

                        with tempfile.NamedTemporaryFile("wb", dir=temp_dir, suffix=ext, delete=False) as temp_file:
                            f.save(temp_file)
                            kwargs["files"].append((f.filename, temp_file.name))
                else:
                    kwargs["files"] = []

            # Now call the function, and validate the output
            try:
                ret = fn(*args, **kwargs)
            finally:
                if temp_dir is not None:
                    shutil.rmtree(temp_dir)

            # Serialize the output it it's not a normal flask response
            if isinstance(ret, Response):
                return ret

            serialized = serialize(ret, accept_type)
            return Response(serialized, content_type=accept_type)

        return wrapper

    return decorate



from typing import Optional, Tuple, Union, Dict, Any

from flask import current_app, g

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.auth import UserInfo, GroupInfo
from qcportal.exceptions import (
    SecurityNotEnabledError,
    UserManagementError,
    AuthorizationFailure,
    AuthenticationFailure,
)

###################################################################
# We have two user endpoints
# 1. /user/<username> is for admins. This allows for changing info
#    about other users, and for enabling/disabling users or changing
#    their roles
# 2. /me allows for changing passwords, fullname, etc of only the
#    logged-in user. Enabling/disabling or changing roles is not
#    allowed through this endpoint, but it is designed to be
#    accessible from all (logged-in users)
#
# Also note that all endpoints are disabled if security is not
# enabled. This is because we don't want to allow modifying users
# while permissions are not checked.
###################################################################


def is_same_user(username_or_id: Union[int, str]) -> bool:
    assert isinstance(username_or_id, (int, str))

    if isinstance(username_or_id, str) and username_or_id.isdecimal():
        username_or_id = int(username_or_id)

    if "user_id" not in g:
        return False
    if "username" not in g:
        return False

    if isinstance(username_or_id, int) and username_or_id == g.user_id:
        return True
    if isinstance(username_or_id, str) and username_or_id == g.username:
        return True

    return False


def assert_security_enabled():
    if not current_app.config["QCFRACTAL_CONFIG"].enable_security:
        raise SecurityNotEnabledError("This functionality is not available if security is not enabled on the server")


# TODO - may not be needed after changing RBAC/ABAC framework
# can be removed after proper/more granular permissions are implemented
def assert_logged_in():
    if g.get("user_id", None) is None:
        raise AuthenticationFailure("Login is required")


def assert_admin():
    if g.get("role", None) != "admin":
        raise AuthorizationFailure("Forbidden: Admin access is required")


#################################
# Groups
#################################


@api_v1.route("/groups", methods=["GET"])
@wrap_global_route("groups", "read", True)
def list_groups_v1():
    assert_security_enabled()
    assert_logged_in()
    return storage_socket.groups.list()


@api_v1.route("/groups", methods=["POST"])
@wrap_global_route("groups", "add", True)
def add_group_v1(body_data: GroupInfo):
    assert_security_enabled()
    assert_logged_in()
    return storage_socket.groups.add(body_data)


@api_v1.route("/groups/<groupname_or_id>", methods=["GET"])
@wrap_global_route("groups", "read", True)
def get_group_v1(groupname_or_id: Union[int, str]):
    assert_security_enabled()
    assert_logged_in()
    return storage_socket.groups.get(groupname_or_id)


@api_v1.route("/groups/<groupname_or_id>", methods=["DELETE"])
@wrap_global_route("groups", "delete", True)
def delete_group_v1(groupname_or_id: Union[int, str]):
    assert_security_enabled()
    assert_logged_in()
    return storage_socket.groups.delete(groupname_or_id)


#################################
# Users
#################################


@api_v1.route("/users", methods=["GET"])
@wrap_global_route("users", "read", True)
def list_users_v1():
    assert_security_enabled()
    assert_logged_in()
    assert_admin()
    return storage_socket.users.list()


@api_v1.route("/users", methods=["POST"])
@wrap_global_route("users", "add", True)
def add_user_v1(body_data: Tuple[UserInfo, Optional[str]]):
    assert_security_enabled()
    assert_logged_in()

    user_info, password = body_data
    return storage_socket.users.add(user_info, password)


@api_v1.route("/users/<username_or_id>", methods=["GET"])
@wrap_global_route("users", "read", True)
def get_user_v1(username_or_id: Union[int, str]):
    assert_security_enabled()
    assert_logged_in()

    # admin can do all
    if g.role == "admin" or is_same_user(username_or_id):
        return storage_socket.users.get(username_or_id)

    raise AuthorizationFailure("Cannot get user information: Forbidden")


@api_v1.route("/me", methods=["GET"])
@wrap_global_route("me", "read", True)
def get_my_user_v1():
    assert_security_enabled()
    assert_logged_in()

    return storage_socket.users.get(g.user_id)


@api_v1.route("/users", methods=["PATCH"])
@wrap_global_route("users", "modify", True)
def modify_user_v1(body_data: UserInfo):
    assert_security_enabled()
    assert_logged_in()

    if g.role == "admin":
        return storage_socket.users.modify(body_data, as_admin=True)

    if body_data.id is None or body_data.username is None:
        raise UserManagementError("Cannot modify user: id or username is missing")

    # Users can only modify themselves, and only certain fields
    # This checks id and name to make sure this user is only modifying themselves
    if is_same_user(body_data.id) and is_same_user(body_data.username):
        return storage_socket.users.modify(body_data, as_admin=False)

    raise AuthorizationFailure("Cannot modify user: Forbidden")


@api_v1.route("/me", methods=["PATCH"])
@wrap_global_route("me", "modify", True)
def modify_my_user_v1(body_data: UserInfo):
    assert_security_enabled()
    assert_logged_in()

    if body_data.id is None or body_data.username is None:
        raise UserManagementError("Cannot modify user: id or username is missing")

    if is_same_user(body_data.id) and is_same_user(body_data.username):
        return storage_socket.users.modify(body_data, as_admin=False)

    raise AuthorizationFailure("Cannot modify user: Forbidden")


@api_v1.route("/users/<username_or_id>/password", methods=["PUT"])
@wrap_global_route("users", "modify", True)
def change_password_v1(username_or_id: Union[int, str], body_data: Optional[str]):
    assert_security_enabled()
    assert_logged_in()

    if g.role == "admin":
        return storage_socket.users.change_password(username_or_id, password=body_data)

    # This checks id or name to make sure this user is only modifying themselves
    if is_same_user(username_or_id):
        return storage_socket.users.change_password(username_or_id, password=body_data)

    raise AuthorizationFailure("Cannot change password: Forbidden")


@api_v1.route("/me/password", methods=["PUT"])
@wrap_global_route("me", "modify", True)
def change_my_password_v1(body_data: Optional[str]):
    assert_security_enabled()
    assert_logged_in()

    return storage_socket.users.change_password(g.user_id, password=body_data)


@api_v1.route("/users/<username_or_id>", methods=["DELETE"])
@wrap_global_route("users", "delete", True)
def delete_user_v1(username_or_id: Union[int, str]):
    assert_security_enabled()
    assert_logged_in()

    if is_same_user(username_or_id):
        raise UserManagementError("Cannot delete your own user")

    return storage_socket.users.delete(username_or_id)


###########################
# User preferences management
###########################
@api_v1.route("/users/<username_or_id>/preferences", methods=["GET"])
@wrap_global_route("users", "read", True)
def get_user_preferences_v1(username_or_id: Union[int, str]):
    assert_security_enabled()
    assert_logged_in()

    if g.role == "admin" or is_same_user(username_or_id):
        return storage_socket.users.get_preferences(g.user_id)
    else:
        raise AuthorizationFailure("Cannot get user preferences: Forbidden")


@api_v1.route("/me/preferences", methods=["GET"])
@wrap_global_route("me", "read", True)
def get_my_preferences_v1():
    assert_security_enabled()
    assert_logged_in()

    return storage_socket.users.get_preferences(g.user_id)


@api_v1.route("/users/<username_or_id>/preferences", methods=["PUT"])
@wrap_global_route("users", "modify", True)
def set_user_preferences_v1(username_or_id: Union[int, str], body_data: Dict[str, Any]):
    assert_security_enabled()
    assert_logged_in()

    if g.role == "admin" or is_same_user(username_or_id):
        return storage_socket.users.set_preferences(g.user_id, body_data)
    else:
        raise AuthorizationFailure("Cannot set user preferences: Forbidden")


@api_v1.route("/me/preferences", methods=["PUT"])
@wrap_global_route("me", "modify", True)
def set_my_preferences_v1(body_data: Dict[str, Any]):
    assert_security_enabled()
    assert_logged_in()

    return storage_socket.users.set_preferences(g.user_id, body_data)


###########################
# User session management
###########################
@api_v1.route("/sessions", methods=["GET"])
@wrap_global_route("users", "read", True)
def list_all_user_sessions_v1():
    assert_security_enabled()
    assert_logged_in()
    assert_admin()

    return storage_socket.auth.list_all_user_sessions()


@api_v1.route("/users/<username_or_id>/sessions", methods=["GET"])
@wrap_global_route("users", "read", True)
def list_user_sessions_v1(username_or_id: Union[int, str]):
    assert_security_enabled()
    assert_logged_in()

    if g.role == "admin" or is_same_user(username_or_id):
        return storage_socket.auth.list_user_sessions(g.user_id)

    raise AuthorizationFailure("Cannot list user sessions: Forbidden")


@api_v1.route("/me/sessions", methods=["GET"])
@wrap_global_route("me", "read", True)
def list_my_sessions_v1():
    assert_security_enabled()
    assert_logged_in()

    return storage_socket.auth.list_user_sessions(g.user_id)



from __future__ import annotations, annotations


def test_serverinfo_motd(snowflake):
    motd_msg = "This is the message of the day" * 10
    client1 = snowflake.client()
    client1.set_motd(motd_msg)

    msg = client1.get_motd()
    assert msg == motd_msg

    client2 = snowflake.client()
    assert client2.server_info["motd"] == motd_msg



"""
Tests the tasks socket (claiming & returning data)
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from qcelemental.models import ComputeError, FailedOperation

from qcfractal.components.managers.db_models import ComputeManagerORM
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.singlepoint.testing_helpers import load_procedure_data, submit_procedure_data
from qcfractalcompute.compress import compress_result
from qcportal.record_models import PriorityEnum, RecordStatusEnum, OutputTypeEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake


def test_task_socket_fullworkflow_success(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    mname, mid = snowflake.activate_manager()
    activated_manager_programs = snowflake.activated_manager_programs()

    id1, result_data1 = submit_procedure_data(storage_socket, "sp_psi4_benzene_energy_1", "tag1", PriorityEnum.normal)
    id2, result_data2 = submit_procedure_data(storage_socket, "sp_psi4_fluoroethane_wfn", "tag1", PriorityEnum.normal)

    time_1 = now_at_utc()

    result_map = {id1: result_data1, id2: result_data2}

    tasks = storage_socket.tasks.claim_tasks(mname.fullname, activated_manager_programs, ["*"])

    with storage_socket.session_scope() as session:
        # Should be claimed in the manager table
        manager = session.get(ComputeManagerORM, mid)
        assert manager.claimed == 2

        # Status should be updated
        for rec_id in [id1, id2]:
            rec = session.get(BaseRecordORM, rec_id)
            assert rec.status == RecordStatusEnum.running
            assert rec.manager_name == mname.fullname
            assert rec.task is not None
            assert rec.compute_history == []
            assert rec.modified_on > time_1
            assert rec.created_on < time_1

        rmeta = storage_socket.tasks.update_finished(
            mname.fullname,
            {
                tasks[0]["id"]: compress_result(result_map[tasks[0]["record_id"]].dict()),
                tasks[1]["id"]: compress_result(result_map[tasks[1]["record_id"]].dict()),
            },
        )

        assert rmeta.n_accepted == 2
        assert rmeta.n_rejected == 0
        assert rmeta.accepted_ids == sorted([id1, id2])

        session.expire_all()
        for rec_id in [id1, id2]:
            rec = session.get(BaseRecordORM, rec_id)
            # Status should be complete
            assert rec.status == RecordStatusEnum.complete

            # Tasks should be deleted
            assert rec.task is None

            # Manager names should have been assigned
            assert rec.manager_name == mname.fullname

            # Modified_on should be updated, but not created_on
            assert rec.created_on < time_1
            assert rec.modified_on > time_1

            # History should have been saved
            assert len(rec.compute_history) == 1
            assert rec.compute_history[0].manager_name == mname.fullname
            assert rec.compute_history[0].modified_on > time_1
            assert rec.compute_history[0].status == RecordStatusEnum.complete

        # Make sure manager info was updated
        manager = session.get(ComputeManagerORM, mid)
        assert manager.successes == 2
        assert manager.failures == 0
        assert manager.rejected == 0
        assert manager.claimed == 2


def test_task_socket_fullworkflow_error(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    mname, mid = snowflake.activate_manager()
    activated_manager_programs = snowflake.activated_manager_programs()

    id1, _ = submit_procedure_data(storage_socket, "sp_psi4_benzene_energy_1")
    id2, _ = submit_procedure_data(storage_socket, "sp_psi4_fluoroethane_wfn")

    time_1 = now_at_utc()

    tasks = storage_socket.tasks.claim_tasks(mname.fullname, activated_manager_programs, ["*"])

    fop = FailedOperation(error=ComputeError(error_type="test_error", error_message="this is a test error"))

    rmeta = storage_socket.tasks.update_finished(
        mname.fullname,
        {tasks[0]["id"]: compress_result(fop.dict()), tasks[1]["id"]: compress_result(fop.dict())},
    )

    assert rmeta.n_accepted == 2
    assert rmeta.n_rejected == 0
    assert rmeta.accepted_ids == sorted([id1, id2])

    with storage_socket.session_scope() as session:
        for rec_id in [id1, id2]:
            rec = session.get(BaseRecordORM, rec_id)

            # Status should be error
            assert rec.status == RecordStatusEnum.error

            # Tasks should not be deleted
            assert rec.task is not None

            # Manager names should have been assigned
            assert rec.manager_name == mname.fullname

            # Modified_on should be updated, but not created_on
            assert rec.created_on < time_1
            assert rec.modified_on > time_1

            # History should have been saved
            assert len(rec.compute_history) == 1
            assert rec.compute_history[0].manager_name == mname.fullname
            assert rec.compute_history[0].modified_on > time_1
            assert rec.compute_history[0].status == RecordStatusEnum.error

        # Make sure manager info was updated
        manager = session.get(ComputeManagerORM, mid)
        assert manager.successes == 0
        assert manager.failures == 2
        assert manager.rejected == 0
        assert manager.claimed == 2


def test_task_socket_fullworkflow_error_retry(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    mname, mid = snowflake.activate_manager()
    activated_manager_programs = snowflake.activated_manager_programs()

    input_spec1, molecule1, result_data1 = load_procedure_data("sp_psi4_benzene_energy_1")
    result_data1_compressed = compress_result(result_data1.dict())

    meta1, id1 = storage_socket.records.singlepoint.add(
        [molecule1], input_spec1, "tag1", PriorityEnum.normal, None, True
    )

    fop = FailedOperation(error=ComputeError(error_type="test_error", error_message="this is a test error"))
    fop_compressed = compress_result(fop.dict())

    # Sends back an error. Do it a few times
    time_0 = now_at_utc()
    tasks = storage_socket.tasks.claim_tasks(mname.fullname, activated_manager_programs, ["*"])
    storage_socket.tasks.update_finished(mname.fullname, {tasks[0]["id"]: fop_compressed})
    storage_socket.records.reset(id1)

    time_1 = now_at_utc()
    tasks = storage_socket.tasks.claim_tasks(mname.fullname, activated_manager_programs, ["*"])
    storage_socket.tasks.update_finished(mname.fullname, {tasks[0]["id"]: fop_compressed})
    storage_socket.records.reset(id1)

    time_2 = now_at_utc()
    tasks = storage_socket.tasks.claim_tasks(mname.fullname, activated_manager_programs, ["*"])
    storage_socket.tasks.update_finished(mname.fullname, {tasks[0]["id"]: fop_compressed})
    storage_socket.records.reset(id1)

    # Now succeed
    time_3 = now_at_utc()
    tasks = storage_socket.tasks.claim_tasks(mname.fullname, activated_manager_programs, ["*"])
    storage_socket.tasks.update_finished(mname.fullname, {tasks[0]["id"]: result_data1_compressed})
    time_4 = now_at_utc()

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, id1)
        hist = rec.compute_history
        assert len(hist) == 4

        for h in hist:
            assert h.manager_name == mname.fullname
            assert len(h.outputs) == 1

        assert time_0 < hist[0].modified_on < time_1
        assert time_1 < hist[1].modified_on < time_2
        assert time_2 < hist[2].modified_on < time_3
        assert time_3 < hist[3].modified_on < time_4

        assert hist[0].status == RecordStatusEnum.error
        assert hist[1].status == RecordStatusEnum.error
        assert hist[2].status == RecordStatusEnum.error
        assert hist[3].status == RecordStatusEnum.complete

        assert list(hist[3].outputs.keys()) == [OutputTypeEnum.stdout]
        assert list(hist[2].outputs.keys()) == [OutputTypeEnum.error]
        assert list(hist[1].outputs.keys()) == [OutputTypeEnum.error]
        assert list(hist[0].outputs.keys()) == [OutputTypeEnum.error]

        # Make sure manager info was updated
        manager = session.get(ComputeManagerORM, mid)
        assert manager.successes == 1
        assert manager.failures == 3
        assert manager.rejected == 0
        assert manager.claimed == 4


def test_task_socket_fullworkflow_error_autoreset(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    activated_manager_programs = snowflake.activated_manager_programs()

    # Change the socket config
    storage_socket.qcf_config.auto_reset.enabled = True
    storage_socket.qcf_config.auto_reset.unknown_error = 1
    storage_socket.qcf_config.auto_reset.random_error = 2

    input_spec1, molecule1, result_data1 = load_procedure_data("sp_psi4_benzene_energy_1")

    meta1, id1 = storage_socket.records.singlepoint.add(
        [molecule1], input_spec1, "tag1", PriorityEnum.normal, None, True
    )

    fop_u = FailedOperation(error=ComputeError(error_type="unknown_error", error_message="this is a test error"))
    fop_u_compressed = compress_result(fop_u.dict())
    fop_r = FailedOperation(error=ComputeError(error_type="random_error", error_message="this is a test error"))
    fop_r_compressed = compress_result(fop_r.dict())

    # Sends back an error
    with storage_socket.session_scope() as session:
        tasks = storage_socket.tasks.claim_tasks(activated_manager_name.fullname, activated_manager_programs, ["*"])
        storage_socket.tasks.update_finished(activated_manager_name.fullname, {tasks[0]["id"]: fop_r_compressed})

        # task should be waiting
        rec = session.get(BaseRecordORM, id1)
        assert rec.status == RecordStatusEnum.waiting
        assert len(rec.compute_history) == 1

        # Claim again, and return a different error
        tasks = storage_socket.tasks.claim_tasks(activated_manager_name.fullname, activated_manager_programs, ["*"])
        storage_socket.tasks.update_finished(activated_manager_name.fullname, {tasks[0]["id"]: fop_u_compressed})

        # waiting again...
        session.expire(rec)
        assert rec.status == RecordStatusEnum.waiting
        assert len(rec.compute_history) == 2

        # Claim again, and return an unknown error. Should stay in errored state now
        tasks = storage_socket.tasks.claim_tasks(activated_manager_name.fullname, activated_manager_programs, ["*"])
        storage_socket.tasks.update_finished(activated_manager_name.fullname, {tasks[0]["id"]: fop_u_compressed})

        session.expire(rec)
        assert rec.status == RecordStatusEnum.error
        assert len(rec.compute_history) == 3



"""merge projects and no_owner branches

Revision ID: e98c86069950
Revises: a036e75160ea, c5a3bed43646
Create Date: 2025-04-11 14:23:21.971882

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "e98c86069950"
down_revision = ("a036e75160ea", "c5a3bed43646")
branch_labels = None
depends_on = None


def upgrade():
    op.drop_constraint("project_owner_group_id_fkey", "project", type_="foreignkey")
    op.drop_constraint("project_owner_user_id_owner_group_id_fkey", "project", type_="foreignkey")
    op.drop_index("ix_project_owner_group_id", table_name="project")
    op.drop_column("project", "owner_group_id")
    pass


def downgrade():
    raise NotImplementedError("Downgrade is not supported")



from __future__ import annotations

import os
from collections import defaultdict
from typing import TYPE_CHECKING

from sqlalchemy import select
from sqlalchemy.orm import selectinload

from qcfractal.components.internal_jobs.status import JobProgress
from qcfractal.components.record_db_models import BaseRecordORM
from qcportal.cache import DatasetCache
from qcportal.dataset_models import BaseDataset
from qcportal.record_models import RecordStatusEnum, BaseRecord
from qcportal.utils import chunk_iterable, make_list

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import Optional, Iterable
    from typing import Iterable
    from sqlalchemy.orm.session import Session


def create_view_file(
    session: Session,
    socket: SQLAlchemySocket,
    dataset_id: int,
    dataset_type: str,
    output_path: str,
    status: Optional[Iterable[RecordStatusEnum]] = None,
    include: Optional[Iterable[str]] = None,
    exclude: Optional[Iterable[str]] = None,
    *,
    include_children: bool = True,
    job_progress: Optional[JobProgress] = None,
):
    """
    Creates a view file for a dataset

    Note: the job progress object will be filled to 90% to leave room for uploading

    Parameters
    ----------
    session
        An existing SQLAlchemy session to use.
    socket
        Full SQLAlchemy socket to use for getting records
    dataset_type
        Type of the underlying dataset to create a view of (as a string)
    dataset_id
        ID of the dataset to create the view for
    output_path
        Full path (including filename) to output the view data to. Must not already exist
    status
        List of statuses to include. Default is to include records with any status
    include
        List of specific record fields to include in the export. Default is to include most fields
    exclude
        List of specific record fields to exclude from the export. Defaults to excluding none.
    include_children
        Specifies whether child records associated with the main records should also be included (recursively)
        in the view file.
    job_progress
        Object used to track the progress of the job
    """

    if os.path.exists(output_path):
        raise RuntimeError(f"File {output_path} exists - will not overwrite")

    if os.path.isdir(output_path):
        raise RuntimeError(f"{output_path} is a directory")

    ds_socket = socket.datasets.get_socket(dataset_type)
    ptl_dataset_type = BaseDataset.get_subclass(dataset_type)

    ptl_entry_type = ptl_dataset_type._entry_type
    ptl_specification_type = ptl_dataset_type._specification_type

    view_db = DatasetCache(output_path, read_only=False, dataset_type=ptl_dataset_type)

    stmt = select(ds_socket.dataset_orm).where(ds_socket.dataset_orm.id == dataset_id)
    stmt = stmt.options(selectinload("*"))
    ds_orm = session.execute(stmt).scalar_one()

    # Metadata
    view_db.update_metadata("dataset_metadata", ds_orm.model_dict())

    # Entries
    if job_progress is not None:
        job_progress.raise_if_cancelled()
        job_progress.update_progress(0, "Processing dataset entries")

    stmt = select(ds_socket.entry_orm)
    stmt = stmt.options(selectinload("*"))
    stmt = stmt.where(ds_socket.entry_orm.dataset_id == dataset_id)

    entries = session.execute(stmt).scalars().all()
    entries = [e.to_model(ptl_entry_type) for e in entries]
    view_db.update_entries(entries)

    if job_progress is not None:
        job_progress.raise_if_cancelled()
        job_progress.update_progress(5, "Processing dataset specifications")

    # Specifications
    stmt = select(ds_socket.specification_orm)
    stmt = stmt.options(selectinload("*"))
    stmt = stmt.where(ds_socket.specification_orm.dataset_id == dataset_id)

    specs = session.execute(stmt).scalars().all()
    specs = [s.to_model(ptl_specification_type) for s in specs]
    view_db.update_specifications(specs)

    if job_progress is not None:
        job_progress.raise_if_cancelled()
        job_progress.update_progress(10, "Loading record information")

    # Now all the records
    stmt = select(ds_socket.record_item_orm).where(ds_socket.record_item_orm.dataset_id == dataset_id)
    stmt = stmt.order_by(ds_socket.record_item_orm.record_id.asc())
    record_items = session.execute(stmt).scalars().all()

    record_ids = set(ri.record_id for ri in record_items)
    all_ids = set(record_ids)

    if include_children:
        # Get all the children ids
        children_ids = socket.records.get_children_ids(session, record_ids)
        all_ids |= set(children_ids)

    # Always include the base record information
    if include is not None:
        include = make_list(include) + ["*"]

    ############################################################################
    # Determine the record types of all the ids (top-level and children if desired)
    ############################################################################
    stmt = select(BaseRecordORM.id, BaseRecordORM.record_type)

    if status is not None:
        stmt = stmt.where(BaseRecordORM.status.in_(status))

    # Sort into a dictionary with keys being the record type
    record_type_map = defaultdict(list)

    for id_chunk in chunk_iterable(all_ids, 500):
        stmt2 = stmt.where(BaseRecordORM.id.in_(id_chunk))
        for record_id, record_type in session.execute(stmt2).yield_per(100):
            record_type_map[record_type].append(record_id)

    if job_progress is not None:
        job_progress.raise_if_cancelled()
        job_progress.update_progress(15, "Processing individual records")

    ############################################################################
    # Actually fetch the record data now
    # We go one over the different types of records, then load them in batches
    ############################################################################
    record_count = len(all_ids)
    finished_count = 0

    for record_type_str, record_ids in record_type_map.items():
        record_socket = socket.records.get_socket(record_type_str)
        record_type = BaseRecord.get_subclass(record_type_str)

        for id_chunk in chunk_iterable(record_ids, 20):
            record_dicts = record_socket.get(id_chunk, include=include, exclude=exclude, session=session)
            record_data = [record_type(**r) for r in record_dicts]
            view_db.update_records(record_data)

            finished_count += len(id_chunk)
            if job_progress is not None:
                job_progress.raise_if_cancelled()

                # Fraction of the 75% left over (15 to start, 10 left over for uploading)
                job_progress.update_progress(
                    15 + int(75 * finished_count / record_count), "Processing individual records"
                )

    # Update the dataset <-> record association
    record_info = [(ri.entry_name, ri.specification_name, ri.record_id) for ri in record_items]
    view_db.update_dataset_records(record_info)



from __future__ import annotations

from typing import Optional, Iterable, Dict, Any

from sqlalchemy import (
    select,
    union,
    Column,
    Integer,
    String,
    JSON,
    Index,
    Computed,
    ForeignKey,
    UniqueConstraint,
    Enum,
)
from sqlalchemy.orm import relationship

from qcfractal.components.auth.db_models import UserIDMapSubquery, UserORM
from qcfractal.components.external_files.db_models import ExternalFileORM
from qcfractal.components.internal_jobs.db_models import InternalJobORM
from qcfractal.db_socket import BaseORM
from qcportal.project_models import ProjectAttachmentType
from qcfractal.db_socket.db_views import view
from qcfractal.components.dataset_db_views import DatasetDirectRecordsView
from qcfractal.components.record_db_views import RecordChildrenView


class ProjectORM(BaseORM):
    """
    Table/ORM for a QCArchive project
    """

    __tablename__ = "project"

    id = Column(Integer, primary_key=True)

    name = Column(String(100), nullable=False)
    lname = Column(String(100), Computed("LOWER(name)"), nullable=False)

    description = Column(String, nullable=False)
    tagline = Column(String, nullable=False)
    tags = Column(JSON, nullable=False)

    default_compute_tag = Column(String, nullable=False)
    default_compute_priority = Column(Integer, nullable=False)

    extras = Column(JSON, nullable=False)

    # Ownership of this project
    owner_user_id = Column(Integer, ForeignKey(UserORM.id), nullable=True)

    owner_user = relationship(
        UserIDMapSubquery,
        foreign_keys=[owner_user_id],
        primaryjoin="ProjectORM.owner_user_id == UserIDMapSubquery.id",
        lazy="selectin",
        viewonly=True,
    )

    attachments = relationship(
        "ProjectAttachmentORM",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    __table_args__ = (
        UniqueConstraint("lname", name="ux_project_project_type_lname"),
        Index("ix_project_owner_user_id", "owner_user_id"),
    )

    _qcportal_model_excludes = ["lname", "owner_user_id"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)

        d["owner_user"] = self.owner_user.username if self.owner_user is not None else None

        return d


class ProjectRecordORM(BaseORM):
    __tablename__ = "project_record"

    project_id = Column(Integer, ForeignKey("project.id", ondelete="cascade"), primary_key=True)
    record_id = Column(Integer, ForeignKey("base_record.id"), nullable=False)

    name = Column(String, nullable=False)
    lname = Column(String(100), Computed("LOWER(name)"), primary_key=True)
    description = Column(String, nullable=False)
    tags = Column(JSON, nullable=False)

    __table_args__ = (Index("ix_project_record_record_id", "record_id"),)


class ProjectDatasetORM(BaseORM):
    __tablename__ = "project_dataset"

    project_id = Column(Integer, ForeignKey("project.id", ondelete="cascade"), primary_key=True)
    dataset_id = Column(Integer, ForeignKey("base_dataset.id"), nullable=False)

    name = Column(String, nullable=False)
    lname = Column(String(100), Computed("LOWER(name)"), primary_key=True)
    description = Column(String, nullable=False)
    tagline = Column(String, nullable=False)
    tags = Column(JSON, nullable=False)

    __table_args__ = (Index("ix_project_dataset_dataset_id", "dataset_id"),)


class ProjectInternalJobORM(BaseORM):
    __tablename__ = "project_internal_job"

    internal_job_id = Column(Integer, ForeignKey(InternalJobORM.id, ondelete="cascade"), primary_key=True)
    project_id = Column(Integer, ForeignKey("project.id", ondelete="cascade"), primary_key=True)


class ProjectAttachmentORM(ExternalFileORM):
    __tablename__ = "project_attachment"

    id = Column(Integer, ForeignKey(ExternalFileORM.id, ondelete="cascade"), primary_key=True)
    project_id = Column(Integer, ForeignKey("project.id", ondelete="cascade"), nullable=False)

    attachment_type = Column(Enum(ProjectAttachmentType), nullable=False)
    tags = Column(JSON, nullable=False)

    __mapper_args__ = {"polymorphic_identity": "project_attachment"}

    __table_args__ = (Index("ix_project_attachment_project_id", "project_id"),)

    _qcportal_model_excludes = ["project_id"]


ProjectRecordsView = view(
    "project_records_view",
    BaseORM.metadata,
    union(
        select(ProjectRecordORM.project_id.label("project_id"), ProjectRecordORM.record_id.label("record_id")),
        select(ProjectRecordORM.project_id.label("project_id"), RecordChildrenView.c.child_id.label("record_id")).join(
            RecordChildrenView, ProjectRecordORM.record_id == RecordChildrenView.c.parent_id
        ),
        select(
            ProjectDatasetORM.project_id.label("project_id"), DatasetDirectRecordsView.c.record_id.label("record_id")
        ).join(DatasetDirectRecordsView, ProjectDatasetORM.dataset_id == DatasetDirectRecordsView.c.dataset_id),
    ),
)



"""Add dataset attachment table

Revision ID: 84285e3620fd
Revises: 02afa97249c7
Create Date: 2025-01-03 10:04:16.201770

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "84285e3620fd"
down_revision = "02afa97249c7"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "dataset_attachment",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("attachment_type", sa.Enum("other", "view", name="datasetattachmenttype"), nullable=False),
        sa.ForeignKeyConstraint(["dataset_id"], ["base_dataset.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(["id"], ["external_file.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index("ix_dataset_attachment_dataset_id", "dataset_attachment", ["dataset_id"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_dataset_attachment_dataset_id", table_name="dataset_attachment")
    op.drop_table("dataset_attachment")
    # ### end Alembic commands ###



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Tuple, Dict, Any

from qcportal.compression import CompressionEnum, compress, decompress
from qcportal.exceptions import MissingDataError
from qcportal.record_models import RecordStatusEnum, OutputTypeEnum
from qcportal.utils import now_at_utc
from .outputstore.utils import create_output_orm
from .record_db_models import (
    RecordComputeHistoryORM,
    BaseRecordORM,
    OutputStoreORM,
    NativeFileORM,
)

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcportal.all_results import AllResultTypes, AllQCPortalRecordTypes, AllSchemaV1ResultTypes
    from typing import Dict, Tuple, List, Any


def build_extras_properties(result: AllResultTypes) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    # Gets rid of numpy arrays
    # Include any of these fields - not all may exist, but pydantic is lenient
    result_dict = result.dict(include={"return_result", "properties", "extras"}, encoding="json")

    new_prop = {}

    return_result = result_dict.get("return_result", None)
    if return_result is not None:
        new_prop["return_result"] = return_result

    extras = result_dict["extras"]

    # Other properties stored in qcvars in extras
    # Store them with the rest of the properties, and remove them from extras
    qcvars = extras.pop("qcvars", {})
    new_prop.update({k.lower(): v for k, v in qcvars.items()})

    properties = result_dict.get("properties", {})
    new_prop.update(properties)

    return extras, new_prop


def upsert_output(session, record_orm: BaseRecordORM, new_output_orm: OutputStoreORM) -> None:
    """
    Insert or replace an output in a records history

    Given a new output orm, if it doesn't exist, add it. If an
    output of the same type already exists, then delete that one and
    insert the new one.
    """
    if len(record_orm.compute_history) == 0:
        raise MissingDataError(f"Record {record_orm.id} does not have any compute history")

    output_type = new_output_orm.output_type
    compute_history = record_orm.compute_history[-1]

    if output_type in compute_history.outputs:
        # TODO - not sure why this is needed. Should be handled by delete-orphan
        old_orm = compute_history.outputs.pop(output_type)
        session.delete(old_orm)
        session.flush()

    compute_history.outputs[output_type] = new_output_orm


def append_output(session: Session, record_orm: BaseRecordORM, output_type: OutputTypeEnum, to_append: str):
    if not to_append:
        return

    if len(record_orm.compute_history) == 0:
        raise MissingDataError(f"Record {record_orm.id} does not have any compute history")

    compute_history = record_orm.compute_history[-1]
    if output_type in compute_history.outputs:
        out_orm = compute_history.outputs[output_type]
        out_str = decompress(out_orm.data, out_orm.compression_type)
        out_str += to_append

        new_data, new_ctype, new_clevel = compress(out_str, CompressionEnum.zstd)
        out_orm.data = new_data
        out_orm.compression_type = new_ctype
        out_orm.compression_level = new_clevel
    else:
        compute_history.outputs[output_type] = create_output_orm(output_type, to_append)

    session.flush()


def compute_history_orms_from_qcportal_record(result: AllQCPortalRecordTypes) -> List[RecordComputeHistoryORM]:
    """
    Retrieves status and (possibly compressed) outputs from a result, and creates
    a record computation history entry
    """

    if not result.compute_history_:
        return []

    history_orms = []

    for ch in result.compute_history_:
        history_orm = RecordComputeHistoryORM()
        history_orm.status = ch.status
        history_orm.manager_name = ch.manager_name
        history_orm.modified_on = ch.modified_on

        if ch.provenance is not None:
            history_orm.provenance = ch.provenance.dict()
        else:
            history_orm.provenance = None

        # Get the compressed outputs if they exist
        if ch.outputs_:
            for output_type, output_store in ch.outputs_.items():
                if output_store.data_ is not None:
                    out_orm = OutputStoreORM(
                        output_type=output_type,
                        compression_type=output_store.compression_type,
                        compression_level=-1,  # We don't know :( But it's not really used anywhere - just for curiosity
                        data=output_store.data_,
                    )

                    history_orm.outputs[output_type] = out_orm

        history_orms.append(history_orm)

    return history_orms


def compute_history_orm_from_schema_v1(result: AllSchemaV1ResultTypes) -> RecordComputeHistoryORM:
    """
    Retrieves status and (possibly compressed) outputs from a result, and creates
    a record computation history entry
    """
    history_orm = RecordComputeHistoryORM()
    history_orm.status = RecordStatusEnum.complete if result.success else RecordStatusEnum.error
    history_orm.provenance = result.provenance.dict()
    history_orm.modified_on = now_at_utc()

    # Get the compressed outputs if they exist
    compressed_output = result.extras.pop("_qcfractal_compressed_outputs", None)

    if compressed_output is not None:
        for output_type, data_dict in compressed_output.items():
            out_orm = OutputStoreORM(
                output_type=output_type,
                compression_type=data_dict["compression_type"],
                compression_level=data_dict["compression_level"],
                data=data_dict["data"],
            )

            history_orm.outputs[output_type] = out_orm

    else:
        if result.stdout is not None:
            stdout_orm = create_output_orm(OutputTypeEnum.stdout, result.stdout)
            history_orm.outputs["stdout"] = stdout_orm
        if result.stderr is not None:
            stderr_orm = create_output_orm(OutputTypeEnum.stderr, result.stderr)
            history_orm.outputs["stderr"] = stderr_orm
        if result.error is not None:
            error_orm = create_output_orm(OutputTypeEnum.error, result.error.dict())
            history_orm.outputs["error"] = error_orm

    return history_orm


def native_files_orms_from_qcportal_record(result: AllQCPortalRecordTypes) -> Dict[str, NativeFileORM]:
    """
    Convert the native files stored in a QCElemental result to an ORM
    """

    if not result.native_files_:
        return {}

    native_files = {}
    for name, nf in result.native_files_.items():
        if nf.data_ is not None:
            nf_orm = NativeFileORM(
                name=name,
                compression_type=nf.compression_type,
                compression_level=-1,  # We don't know :( But it's not really used anywhere - just for curiosity
                data=nf.data_,
            )
            native_files[name] = nf_orm

    return native_files


def native_files_orms_from_schema_v1(result: AllSchemaV1ResultTypes) -> Dict[str, NativeFileORM]:
    """
    Convert the native files stored in a QCElemental result to an ORM
    """

    compressed_nf = result.extras.pop("_qcfractal_compressed_native_files", None)

    if compressed_nf is not None:
        native_files = {}
        for name, nf_data in compressed_nf.items():
            # nf_data is a dictionary with keys 'data', 'compression_type', "compression_level"
            nf_orm = NativeFileORM(
                name=name,
                compression_type=nf_data["compression_type"],
                compression_level=nf_data["compression_level"],
                data=nf_data["data"],
            )
            native_files[name] = nf_orm

        return native_files
    elif "native_files" in result.__fields__:  # Not compressed, but part of result
        native_files = {}
        for name, nf_data in result.native_files.items():

            compressed_data, compression_type, compression_level = compress(nf_data, CompressionEnum.zstd)
            nf_orm = NativeFileORM(
                name=name,
                compression_type=compression_type,
                compression_level=compression_level,
                data=compressed_data,
            )

            native_files[name] = nf_orm

        return native_files
    else:
        return {}



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

import pytest

from qcarchivetesting import load_molecule_data
from qcportal.optimization import (
    OptimizationSpecification,
)
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.singlepoint import QCSpecification
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcportal import PortalClient

from qcfractal.components.optimization.testing_helpers import (
    compare_optimization_specs,
    test_specs,
    submit_procedure_data,
    run_procedure_data,
)


def test_optimization_client_tag_priority(snowflake_client: PortalClient):
    water = load_molecule_data("water_dimer_minima")

    for tag, priority in itertools.product(["*", "tag99"], list(PriorityEnum)):
        meta1, id1 = snowflake_client.add_optimizations(
            [water],
            "prog",
            QCSpecification(
                program="prog",
                method="hf",
                basis="sto-3g",
                driver="deferred",
                keywords={"tag_priority": [tag, priority]},
            ),
            compute_priority=priority,
            compute_tag=tag,
        )

        assert meta1.n_inserted == 1
        rec = snowflake_client.get_records(id1, include=["task"])
        assert rec[0].task.compute_tag == tag
        assert rec[0].task.compute_priority == priority


@pytest.mark.parametrize("spec", test_specs)
def test_optimization_client_add_get(submitter_client: PortalClient, spec: OptimizationSpecification):
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")
    ne4 = load_molecule_data("neon_tetramer")
    all_mols = [water, hooh, ne4]

    time_0 = now_at_utc()
    meta, id = submitter_client.add_optimizations(
        initial_molecules=all_mols,
        program=spec.program,
        keywords=spec.keywords,
        protocols=spec.protocols,
        qc_specification=spec.qc_specification,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
    )

    time_1 = now_at_utc()
    assert meta.success

    recs = submitter_client.get_optimizations(id, include=["task", "initial_molecule"])

    assert len(recs) == 3

    for r in recs:
        assert r.record_type == "optimization"
        assert r.record_type == "optimization"
        assert compare_optimization_specs(spec, r.specification)

        assert r.status == RecordStatusEnum.waiting
        assert r.children_status == {}

        assert r.task.function is None
        assert r.task.compute_tag == "tag1"
        assert r.task.compute_priority == PriorityEnum.low

        assert r.creator_user == submitter_client.username

        assert time_0 < r.created_on < time_1
        assert time_0 < r.modified_on < time_1

    mol1 = submitter_client.get_molecules([recs[0].initial_molecule_id])[0]
    mol2 = submitter_client.get_molecules([recs[1].initial_molecule_id])[0]
    mol3 = submitter_client.get_molecules([recs[2].initial_molecule_id])[0]
    assert mol1.identifiers.molecule_hash == water.get_hash()
    assert recs[0].initial_molecule.identifiers.molecule_hash == water.get_hash()

    assert mol2.identifiers.molecule_hash == hooh.get_hash()
    assert recs[1].initial_molecule.identifiers.molecule_hash == hooh.get_hash()

    assert mol3.identifiers.molecule_hash == ne4.get_hash()
    assert recs[2].initial_molecule.identifiers.molecule_hash == ne4.get_hash()


@pytest.mark.parametrize("spec", test_specs)
@pytest.mark.parametrize("find_existing", [True, False])
def test_optimization_client_add_duplicate(
    submitter_client: PortalClient, spec: OptimizationSpecification, find_existing: bool
):
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")
    ne4 = load_molecule_data("neon_tetramer")
    all_mols = [water, hooh, ne4]

    meta, id = submitter_client.add_optimizations(
        initial_molecules=all_mols,
        program=spec.program,
        keywords=spec.keywords,
        protocols=spec.protocols,
        qc_specification=spec.qc_specification,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
        find_existing=True,
    )

    assert meta.success
    assert meta.n_inserted == len(all_mols)

    meta, id2 = submitter_client.add_optimizations(
        initial_molecules=all_mols,
        program=spec.program,
        keywords=spec.keywords,
        protocols=spec.protocols,
        qc_specification=spec.qc_specification,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
        find_existing=find_existing,
    )

    if find_existing:
        assert meta.n_existing == len(all_mols)
        assert meta.n_inserted == 0
        assert id == id2
    else:
        assert meta.n_existing == 0
        assert meta.n_inserted == len(all_mols)
        assert set(id).isdisjoint(id2)


def test_optimization_client_add_existing_molecule(snowflake_client: PortalClient):
    spec = test_specs[0]

    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")
    ne4 = load_molecule_data("neon_tetramer")
    all_mols = [water, hooh, ne4]

    # Add a molecule separately
    _, mol_ids = snowflake_client.add_molecules([ne4])

    # Now add records
    meta, id = snowflake_client.add_optimizations(
        initial_molecules=all_mols,
        program=spec.program,
        keywords=spec.keywords,
        protocols=spec.protocols,
        qc_specification=spec.qc_specification,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
    )
    recs = snowflake_client.get_optimizations(id)

    assert len(recs) == 3
    assert recs[2].initial_molecule_id == mol_ids[0]


@pytest.mark.parametrize("opt_file", ["opt_psi4_benzene", "opt_psi4_fluoroethane_notraj"])
def test_optimization_client_delete(snowflake: QCATestingSnowflake, opt_file: str):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    opt_id = run_procedure_data(storage_socket, activated_manager_name, opt_file)

    rec = snowflake_client.get_optimizations(opt_id)
    child_ids = [x.id for x in rec.trajectory]

    meta = snowflake_client.delete_records(opt_id, soft_delete=True, delete_children=False)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == 0

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x.status == RecordStatusEnum.complete for x in child_recs)
    opt_rec = snowflake_client.get_records(opt_id)
    if child_ids:
        assert opt_rec.children_status == {RecordStatusEnum.complete: len(child_ids)}

    # Undo what we just did
    snowflake_client.undelete_records(opt_id)

    meta = snowflake_client.delete_records(opt_id, soft_delete=True, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == len(child_ids)

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x.status == RecordStatusEnum.deleted for x in child_recs)
    opt_rec = snowflake_client.get_records(opt_id)
    if child_ids:
        assert opt_rec.children_status == {RecordStatusEnum.deleted: len(child_ids)}

    meta = snowflake_client.delete_records(opt_id, soft_delete=False, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == len(child_ids)

    recs = snowflake_client.get_optimizations(opt_id, missing_ok=True)
    assert recs is None

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x is None for x in child_recs)

    # DB should be pretty empty now
    query_res = snowflake_client.query_records()
    query_res_l = list(query_res)
    assert len(query_res_l) == 0


def test_optimization_client_harddelete_nochildren(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    opt_id = run_procedure_data(storage_socket, activated_manager_name, "opt_psi4_benzene")

    rec = snowflake_client.get_optimizations(opt_id)
    child_ids = [x.id for x in rec.trajectory]

    meta = snowflake_client.delete_records(opt_id, soft_delete=False, delete_children=False)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == 0

    recs = snowflake_client.get_optimizations(opt_id, missing_ok=True)
    assert recs is None

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x is not None for x in child_recs)


@pytest.mark.parametrize("opt_file", ["opt_psi4_benzene", "opt_psi4_methane_sometraj"])
def test_optimization_client_delete_traj_inuse(snowflake: QCATestingSnowflake, opt_file: str):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    opt_id = run_procedure_data(storage_socket, activated_manager_name, opt_file)

    rec = snowflake_client.get_optimizations(opt_id)
    child_ids = [x.id for x in rec.trajectory]

    meta = snowflake_client.delete_records(child_ids[0], soft_delete=False)
    assert meta.success is False
    assert meta.error_idx == [0]

    ch_rec = snowflake_client.get_records(child_ids[0])
    assert ch_rec is not None


@pytest.mark.parametrize("opt_file", ["opt_psi4_benzene", "opt_psi4_methane_sometraj"])
@pytest.mark.parametrize("fetch_traj", [True, False])
def test_optimization_client_traj(snowflake: QCATestingSnowflake, opt_file: str, fetch_traj: bool):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    opt_id = run_procedure_data(storage_socket, activated_manager_name, opt_file)

    rec = snowflake_client.get_optimizations(opt_id)
    rec_traj = snowflake_client.get_optimizations(opt_id, include=["trajectory"])

    assert rec_traj.trajectory is not None

    if fetch_traj:
        rec._fetch_trajectory()
        assert rec.trajectory_ids_ is not None
        assert rec.trajectory_records_ is not None
    else:
        assert rec.trajectory_ids_ is None
        assert rec.trajectory_records_ is None

    assert rec.trajectory_element(0).id == rec_traj.trajectory[0].id
    assert rec.trajectory_element(-1).id == rec_traj.trajectory[-1].id


def test_optimization_client_query(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    id_1, _ = submit_procedure_data(storage_socket, "opt_psi4_fluoroethane_notraj")
    id_2, _ = submit_procedure_data(storage_socket, "opt_psi4_benzene")
    id_3, _ = submit_procedure_data(storage_socket, "opt_psi4_methane_sometraj")

    recs = snowflake_client.get_optimizations([id_1, id_2, id_3])

    # query for molecule
    query_res = snowflake_client.query_optimizations(initial_molecule_id=[recs[1].initial_molecule_id])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    # query for program
    query_res = snowflake_client.query_optimizations(program=["psi4"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    # query for program
    query_res = snowflake_client.query_optimizations(program=["geometric"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 3

    query_res = snowflake_client.query_optimizations(qc_program=["psi4"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 3

    # query for basis
    query_res = snowflake_client.query_optimizations(qc_basis=["sTO-3g"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_optimizations(qc_basis=[None])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_optimizations(qc_basis=[""])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    # query for method
    query_res = snowflake_client.query_optimizations(qc_method=["b3lyP"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 3

    # Some empty queries
    query_res = snowflake_client.query_optimizations(program=["madeupprog"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    # Query by default returns everything
    query_res = snowflake_client.query_optimizations()
    query_res_l = list(query_res)
    assert len(query_res_l) == 3

    # Query by default (with a limit)
    query_res = snowflake_client.query_optimizations(limit=1)
    query_res_l = list(query_res)
    assert len(query_res_l) == 1



import pytest


@pytest.fixture(scope="function", params=[True, False])
def spec_test_runner(storage_socket, request):

    def _run_test(record_type, spec_1, spec_2, expected_same: bool):
        socket = storage_socket.records.get_socket(record_type)
        if request.param:
            meta, ids = socket.add_specifications([spec_1, spec_2, spec_1, spec_2, spec_1, spec_1, spec_2])
            assert meta.success
            if expected_same:
                assert meta.inserted_idx == [0]
                assert meta.existing_idx == [1, 2, 3, 4, 5, 6]
                assert all(x == ids[0] for x in ids)
            else:
                assert meta.inserted_idx == [0, 1]
                assert meta.existing_idx == [2, 3, 4, 5, 6]
                assert ids[0] == ids[2]
                assert ids[0] == ids[4]
                assert ids[0] == ids[5]

                assert ids[1] == ids[3]
                assert ids[1] == ids[6]

                assert ids[0] != ids[1]

            # Try adding again
            meta, ids = socket.add_specifications([spec_1, spec_2, spec_1, spec_2, spec_1, spec_1, spec_2])
            assert meta.success
            assert meta.inserted_idx == []
            assert meta.existing_idx == [0, 1, 2, 3, 4, 5, 6]
            if expected_same:
                assert all(x == ids[0] for x in ids)
            else:
                assert ids[0] == ids[2]
                assert ids[0] == ids[4]
                assert ids[0] == ids[5]

                assert ids[1] == ids[3]
                assert ids[1] == ids[6]

                assert ids[0] != ids[1]

        else:
            meta1, id1 = socket.add_specification(spec_1)
            meta2, id2 = socket.add_specification(spec_2)

            assert meta1.success
            assert meta2.success

            assert meta1.inserted_idx == [0]
            assert meta2.inserted_idx == ([] if expected_same else [0])
            assert meta2.existing_idx == ([0] if expected_same else [])
            assert id1 == id2 if expected_same else id1 != id2

    return _run_test



"""Add dataset internal jobs

Revision ID: 5f6f804e11d3
Revises: 84285e3620fd
Create Date: 2025-01-09 16:25:50.187495

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "5f6f804e11d3"
down_revision = "84285e3620fd"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "dataset_internal_job",
        sa.Column("internal_job_id", sa.Integer(), nullable=False),
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["dataset_id"], ["base_dataset.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(["internal_job_id"], ["internal_jobs.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("internal_job_id", "dataset_id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("dataset_internal_job")
    # ### end Alembic commands ###



from __future__ import annotations

import inspect
import logging
import select as io_select
import traceback
import uuid
from datetime import datetime, timedelta
from operator import attrgetter
from socket import gethostname
from typing import TYPE_CHECKING

import psycopg2.extensions
from sqlalchemy import select, delete, update, and_, or_
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.exc import IntegrityError

from qcfractal.components.auth.db_models import UserIDMapSubquery
from qcfractal.db_socket.helpers import get_query_proj_options
from qcportal.exceptions import MissingDataError
from qcportal.internal_jobs.models import InternalJobStatusEnum, InternalJobQueryFilters
from qcportal.serialization import encode_to_json
from qcportal.utils import now_at_utc
from .db_models import InternalJobORM
from .status import JobProgress, CancelledJobException, JobRunnerStoppingException

if TYPE_CHECKING:
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import Optional, Dict, Any
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import Dict, Optional, Any, List

_default_error = {"error_type": "not_supplied", "error_message": "No error message found on task."}


class InternalJobSocket:
    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

        self._hostname = gethostname()

        # How often to update progress (in seconds)
        # Hardcoded for now
        self._update_frequency = 5

        # Old internal job cleanup
        self._delete_internal_job_frequency = 60 * 60 * 24  # one day
        self._internal_job_keep = root_socket.qcf_config.internal_job_keep

        if self._internal_job_keep > 0:
            with self.root_socket.session_scope() as session:
                self.add(
                    "delete_old_internal_jobs",
                    now_at_utc() + timedelta(seconds=2.0),
                    "internal_jobs.delete_old_internal_jobs",
                    {},
                    user_id=None,
                    unique_name=True,
                    repeat_delay=self._delete_internal_job_frequency,
                    session=session,
                )

    def add(
        self,
        name: str,
        scheduled_date: datetime,
        function: str,
        kwargs: Dict[str, Any],
        user_id: Optional[int],
        unique_name: bool = False,
        after_function: Optional[str] = None,
        after_function_kwargs: Optional[Dict[str, Any]] = None,
        repeat_delay: Optional[int] = None,
        serial_group: Optional[str] = None,
        *,
        session: Optional[Session] = None,
    ) -> int:
        """
        Adds a new job to the internal job queue

        Parameters
        ----------
        name
            Descriptive name of the job
        scheduled_date
            When the job should be run. If the server is not running when this time elapses, it will
            run when it comes back up.
        function
            The function to run, as a string. Should be a member of the sqlalchemy socket.
            Example: `services.iterate_services`
        kwargs
            Arguments to pass to the function
        user_id
            The user making creating this job
        unique_name
            If true, do not add if a job with that name already exists in the job queue.
        after_function
            When this job is done, call this function
        after_function_kwargs
            Arguments to use when calling `after_function`
        repeat_delay
            If set, will submit a new, identical job to be run repeat_delay seconds after this one finishes
        serial_group
            Only one job within this group may be run at once. If None, there is no limit
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A unique ID representing this job in the queue
        """

        with self.root_socket.optional_session(session) as session:
            if unique_name:
                stmt = insert(InternalJobORM)
                stmt = stmt.values(
                    name=name,
                    status=InternalJobStatusEnum.waiting,
                    unique_name=name,
                    scheduled_date=scheduled_date,
                    function=function,
                    kwargs=kwargs,
                    after_function=after_function,
                    after_function_kwargs=after_function_kwargs,
                    repeat_delay=repeat_delay,
                    serial_group=serial_group,
                    user_id=user_id,
                )
                stmt = stmt.on_conflict_do_update(
                    constraint="ux_internal_jobs_unique_name",
                    set_={
                        "after_function": after_function,
                        "after_function_kwargs": after_function_kwargs,
                        "repeat_delay": repeat_delay,
                    },
                )
                stmt = stmt.returning(InternalJobORM.id)
                job_id = session.execute(stmt).scalar_one_or_none()

                if job_id is not None:
                    self._logger.debug(f"Job with name {name} added or updated - id: {job_id}")
                else:
                    # Nothing was returned, meaning nothing was inserted
                    self._logger.debug(f"Job with name {name} already found. Not adding?")
                    stmt = select(InternalJobORM.id).where(InternalJobORM.unique_name == name)
                    job_id = session.execute(stmt).scalar_one_or_none()

            else:
                job_orm = InternalJobORM(
                    name=name,
                    status=InternalJobStatusEnum.waiting,
                    scheduled_date=scheduled_date,
                    function=function,
                    kwargs=kwargs,
                    after_function=after_function,
                    after_function_kwargs=after_function_kwargs,
                    repeat_delay=repeat_delay,
                    serial_group=serial_group,
                    user_id=user_id,
                )
                if unique_name:
                    job_orm.unique_name = name

                session.add(job_orm)
                session.flush()
                job_id = job_orm.id

        return job_id

    def get(self, job_id: int, *, session: Optional[Session] = None) -> Dict[str, Any]:
        """
        Obtain all the information about a job

        An exception is raised if a job with the given ID doesn't exist

        Parameters
        ----------
        job_id
            ID representing the job
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Information about the job (as a dictionary)
        """
        stmt = select(InternalJobORM).where(InternalJobORM.id == job_id)
        with self.root_socket.optional_session(session) as session:
            job_orm = session.execute(stmt).scalar_one_or_none()
            if job_orm is None:
                raise MissingDataError(f"Internal job with id={job_id} not found")

            return job_orm.model_dict()

    def query(self, query_data: InternalJobQueryFilters, *, session: Optional[Session] = None) -> List[Dict[str, Any]]:
        """
        General query of internal jobs in the database

        All search criteria are merged via 'and'. Therefore, records will only
        be found that match all the criteria.

        Parameters
        ----------
        query_data
            Fields/filters to query for
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A list of job info (as dictionaries) that were found in the database.
        """

        proj_options = get_query_proj_options(InternalJobORM, query_data.include, query_data.exclude)

        stmt = select(InternalJobORM)

        and_query = []
        if query_data.job_id is not None:
            and_query.append(InternalJobORM.id.in_(query_data.job_id))
        if query_data.name is not None:
            and_query.append(InternalJobORM.name.in_(query_data.name))
        if query_data.runner_hostname is not None:
            and_query.append(InternalJobORM.runner_hostname.in_(query_data.runner_hostname))
        if query_data.status is not None:
            and_query.append(InternalJobORM.status.in_(query_data.status))
        if query_data.last_updated_before is not None:
            and_query.append(InternalJobORM.last_updated < query_data.last_updated_before)
        if query_data.last_updated_after is not None:
            and_query.append(InternalJobORM.last_updated > query_data.last_updated_after)
        if query_data.added_before is not None:
            and_query.append(InternalJobORM.added_date < query_data.added_before)
        if query_data.added_after is not None:
            and_query.append(InternalJobORM.added_date > query_data.added_after)
        if query_data.scheduled_before is not None:
            and_query.append(InternalJobORM.scheduled_date < query_data.scheduled_before)
        if query_data.scheduled_after is not None:
            and_query.append(InternalJobORM.scheduled_date > query_data.scheduled_after)
        if query_data.user:
            stmt = stmt.join(UserIDMapSubquery)

            int_ids = {x for x in query_data.user if isinstance(x, int) or x.isdecimal()}
            str_names = set(query_data.user) - int_ids

            and_query.append(or_(UserIDMapSubquery.username.in_(str_names), UserIDMapSubquery.id.in_(int_ids)))

        with self.root_socket.optional_session(session, True) as session:
            stmt = stmt.filter(and_(True, *and_query))
            stmt = stmt.options(*proj_options)

            if query_data.cursor is not None:
                stmt = stmt.where(InternalJobORM.id < query_data.cursor)

            stmt = stmt.order_by(InternalJobORM.id.desc())
            stmt = stmt.limit(query_data.limit)
            stmt = stmt.distinct(InternalJobORM.id)

            results = session.execute(stmt).scalars().all()
            result_dicts = [x.model_dict() for x in results]

        return result_dicts

    def delete(self, job_id: int, *, session: Optional[Session] = None):
        """
        Delete a job from the job queue

        Parameters
        ----------
        job_id
            ID representing the job
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        stmt = delete(InternalJobORM).where(InternalJobORM.id == job_id)
        with self.root_socket.optional_session(session) as session:
            session.execute(stmt)

    def delete_old_internal_jobs(self, session: Session) -> None:
        """
        Deletes old internal jobs (as defined by the configuration)
        """

        # we check when adding the job, but double check here
        if self._internal_job_keep <= 0:
            return

        before = now_at_utc() - timedelta(seconds=self._internal_job_keep)

        stmt = delete(InternalJobORM)
        stmt = stmt.where(
            InternalJobORM.status.in_(
                (InternalJobStatusEnum.complete, InternalJobStatusEnum.error, InternalJobStatusEnum.cancelled)
            )
        )
        stmt = stmt.where(InternalJobORM.ended_date < before)
        r = session.execute(stmt)
        num_deleted = r.rowcount
        self._logger.info(f"Deleted {num_deleted} internal jobs before {before}")

    def cancel(self, job_id: int, *, session: Optional[Session] = None):
        """
        Cancels a job in the job queue

        Parameters
        ----------
        job_id
            ID representing the job
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        cancellable = [InternalJobStatusEnum.waiting, InternalJobStatusEnum.running]
        stmt = update(InternalJobORM)
        stmt = stmt.where(InternalJobORM.id == job_id)
        stmt = stmt.where(InternalJobORM.status.in_(cancellable))
        stmt = stmt.values(status=InternalJobStatusEnum.cancelled)

        with self.root_socket.optional_session(session) as session:
            session.execute(stmt)

    def _run_single(self, session: Session, job_orm: InternalJobORM, logger, job_progress: JobProgress):
        """
        Runs a single job
        """

        # For logging (ORM may end up detached or somthing)
        job_id = job_orm.id

        try:
            func_attr = attrgetter(job_orm.function)

            # Function must be part of the sockets
            func = func_attr(self.root_socket)

            # We need to determine the parameters of this function
            func_params = inspect.signature(func).parameters

            add_kwargs = {}

            # If the function has a "job_progress" and/or "session" args, pass those in
            if "job_progress" in func_params:
                add_kwargs["job_progress"] = job_progress

            if "session" in func_params:
                add_kwargs["session"] = session

            # Run the desired function
            # Raises an exception if cancelled
            result = func(**job_orm.kwargs, **add_kwargs)

            job_orm.status = InternalJobStatusEnum.complete
            job_orm.progress = 100
            job_orm.progress_description = "Complete"
            job_orm.result = encode_to_json(result)

        # Job itself is being cancelled
        except CancelledJobException:
            session.rollback()
            logger.info(f"Job {job_id} was cancelled")
            job_orm.status = InternalJobStatusEnum.cancelled
            job_orm.result = None

        # Runner is stopping down
        except JobRunnerStoppingException:
            session.rollback()
            logger.info(f"Job {job_id} was running, but runner is stopping")

            # Basically reset everything
            job_orm.status = InternalJobStatusEnum.waiting
            job_orm.progress = 0
            job_orm.progress_description = None
            job_orm.started_date = None
            job_orm.last_updated = None
            job_orm.result = None
            job_orm.runner_uuid = None

        # Function itself had an error
        except:
            session.rollback()
            job_orm.result = traceback.format_exc()
            logger.error(f"Job {job_id} failed with exception:\n{job_orm.result}")
            job_orm.status = InternalJobStatusEnum.error

        if job_progress.deleted:
            # Row does not exist anymore
            session.expunge(job_orm)
        else:
            # If status is waiting, that means the runner itself is stopping or something
            if job_orm.status != InternalJobStatusEnum.waiting:
                job_orm.ended_date = now_at_utc()
                job_orm.last_updated = job_orm.ended_date

                # Clear the unique name so we can add another one if needed
                has_unique_name = job_orm.unique_name is not None
                job_orm.unique_name = None

            # Flush but don't commit. This will prevent marking the task as finished
            # before the after_func has been run, but allow new ones to be added
            # with unique_name = True
            session.flush()

            # Run the function specified to be run after
            if job_orm.status == InternalJobStatusEnum.complete and job_orm.after_function is not None:
                try:
                    after_func_attr = attrgetter(job_orm.after_function)
                    after_func = after_func_attr(self.root_socket)

                    after_func_params = inspect.signature(after_func).parameters
                    add_after_kwargs = {}
                    if "session" in after_func_params:
                        add_after_kwargs["session"] = session
                    after_func(**job_orm.after_function_kwargs, **add_after_kwargs)
                except Exception:
                    # Don't rollback? not sure what to do here
                    result = traceback.format_exc()
                    logger.error(f"Job {job_orm.id} failed with exception:\n{result}")

                    job_orm.status = InternalJobStatusEnum.error

            if job_orm.status == InternalJobStatusEnum.complete and job_orm.repeat_delay is not None:
                self.add(
                    name=job_orm.name,
                    scheduled_date=now_at_utc() + timedelta(seconds=job_orm.repeat_delay),
                    function=job_orm.function,
                    kwargs=job_orm.kwargs,
                    user_id=job_orm.user_id,
                    unique_name=has_unique_name,
                    after_function=job_orm.after_function,
                    after_function_kwargs=job_orm.after_function_kwargs,
                    repeat_delay=job_orm.repeat_delay,
                    session=session,
                )

        session.commit()

    @staticmethod
    def _wait_for_job(session: Session, logger, conn, end_event):
        """
        Blocks until a job is possibly available to run
        """

        serial_cte = select(InternalJobORM.serial_group).distinct()
        serial_cte = serial_cte.where(InternalJobORM.status == InternalJobStatusEnum.running)
        serial_cte = serial_cte.where(InternalJobORM.serial_group.is_not(None))
        serial_cte = serial_cte.cte()

        next_job_stmt = select(InternalJobORM.scheduled_date)
        next_job_stmt = next_job_stmt.where(InternalJobORM.status == InternalJobStatusEnum.waiting)
        next_job_stmt = next_job_stmt.where(
            or_(InternalJobORM.serial_group.is_(None), InternalJobORM.serial_group.not_in(select(serial_cte)))
        )
        next_job_stmt = next_job_stmt.order_by(InternalJobORM.scheduled_date.asc())

        # Skip any that are being claimed for running right now
        next_job_stmt = next_job_stmt.with_for_update(skip_locked=True, read=True)
        next_job_stmt = next_job_stmt.limit(1)

        cursor = conn.cursor()

        # Start listening for notifications
        cursor.execute("LISTEN check_internal_jobs;")

        while True:
            # Remove any pending notifications
            conn.poll()
            conn.notifies.clear()

            # Find the next available job, and find out if we have to sleep until then
            next_job_time = session.execute(next_job_stmt).scalar_one_or_none()
            now = now_at_utc()
            if next_job_time is None:
                # Wait up to 5 minutes by default. This is just a catch all to prevent
                # programming mistakes from causing infinite waits
                total_to_wait = 300.0
            else:
                total_to_wait = (next_job_time - now).total_seconds()

            session.rollback()  # Release the transaction (and row level lock)

            # If this is <= 0, we don't have to wait
            if total_to_wait <= 0.0:
                logger.debug("not waiting, found possible job to run")
                break

            logger.debug(f"found future job scheduled for {next_job_time}, waiting up to {total_to_wait:.2f} seconds")

            # This will end either if we have waited long enough, or there
            # is a notification from postgres (this connection has run LISTEN)
            total_waited = 0.0

            # Wait in 2 second intervals (to check for end_event)
            while total_waited < total_to_wait:
                to_wait = min(total_to_wait - total_waited, 2.0)
                if to_wait <= 0.0:
                    break

                # waits until a notification is received, up to 5 seconds
                # https://docs.python.org/3/library/select.html#select.select
                try:
                    if io_select.select([conn], [], [], to_wait) != ([], [], []):
                        # We got a notification
                        logger.debug("received notification from check_internal_jobs")
                        conn.poll()

                        # We don't actually care about the individual notifications
                        conn.notifies.clear()

                        # Go back to the outer loop
                        break
                    else:
                        # select timed out. Check for event (we should shut down)
                        if end_event.is_set():
                            break
                        total_waited += to_wait
                except KeyboardInterrupt:
                    break

            if end_event.is_set():
                break

        # Stop listening for insertions on internal_jobs
        cursor.execute("UNLISTEN check_internal_jobs;")
        cursor.close()

    def run_loop(self, end_event):
        """
        Runs in a infinite loop, checking for jobs and running them

        Parameters
        ----------
        end_event
            An event (threading.Event, multiprocessing.Event) that, when set, will
            stop this loop
        """

        # give this loop a unique uuid
        runner_uuid = str(uuid.uuid4())

        # Get a uuid-specific logger
        logger = logging.getLogger(f"internal_job_runner:{runner_uuid}")

        # Two sessions - one for the object, and one for the job status object
        session_main = self.root_socket.Session()
        session_status = self.root_socket.Session()

        # Set up the listener for postgres. This will be notified when something is
        # added to the internal job queue
        # We use a raw psycopg2 connection; sqlalchemy doesn't directly support LISTEN/NOTIFY
        conn = self.root_socket.engine.raw_connection()
        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)

        # Prepare a statement for finding jobs. Filters will be added in the loop
        serial_cte = select(InternalJobORM.serial_group).distinct()
        serial_cte = serial_cte.where(InternalJobORM.status == InternalJobStatusEnum.running)
        serial_cte = serial_cte.where(InternalJobORM.serial_group.is_not(None))
        serial_cte = serial_cte.cte()

        stmt = select(InternalJobORM)
        stmt = stmt.order_by(InternalJobORM.scheduled_date.asc()).limit(1)
        stmt = stmt.with_for_update(skip_locked=True)

        while True:
            if end_event.is_set():
                logger.info("shutting down")
                break

            logger.debug("checking for jobs")

            # Pick up anything waiting, or anything that hasn't been updated in a while (12 update periods)

            now = now_at_utc()
            dead = now - timedelta(seconds=(self._update_frequency * 12))
            logger.debug(f"checking for jobs before date {now}")

            # Job is waiting, schedule to be run now or in the past,
            # Serial group does not have any running or is not set
            cond1 = and_(
                InternalJobORM.status == InternalJobStatusEnum.waiting,
                InternalJobORM.scheduled_date <= now,
                or_(InternalJobORM.serial_group.is_(None), InternalJobORM.serial_group.not_in(select(serial_cte))),
            )

            # Job is running but runner is determined to be dead. Serial group doesn't matter
            cond2 = and_(InternalJobORM.status == InternalJobStatusEnum.running, InternalJobORM.last_updated < dead)

            stmt_now = stmt.where(or_(cond1, cond2))
            job_orm = session_main.execute(stmt_now).scalar_one_or_none()

            # If no job was found, wait for one
            if job_orm is None:
                session_main.rollback()  # release the transaction
                logger.debug("no jobs found")
                self._wait_for_job(session_main, logger, conn, end_event)

                if end_event.is_set():
                    logger.info("shutting down")
                    break
                else:
                    continue

            if end_event.is_set():
                logger.info("shutting down")
                break

            logger.info(f"running job {job_orm.name} id={job_orm.id} scheduled_date={job_orm.scheduled_date}")
            job_orm.started_date = now_at_utc()
            job_orm.last_updated = now_at_utc()
            job_orm.runner_hostname = self._hostname
            job_orm.runner_uuid = runner_uuid
            job_orm.status = InternalJobStatusEnum.running

            # For logging below - the object might be in an odd state after the exception, where accessing
            # it results in further exceptions
            serial_group = job_orm.serial_group

            # Violation of the unique constraint may occur - two runners attempting to take
            # different jobs of the same serial group at the same time
            try:
                # Releases the row-level lock (from the with_for_update() in the original query)
                session_main.commit()
            except IntegrityError:
                logger.info(
                    f"Attempting to run job from serial group '{serial_group}, but seems like another runner got to another job from the same group first"
                )
                session_main.rollback()
                continue

            job_progress = JobProgress(job_orm.id, runner_uuid, session_status, self._update_frequency, end_event)
            self._run_single(session_main, job_orm, logger, job_progress=job_progress)

            # Stop the updating thread and cleanup
            job_progress.stop()

        session_main.close()
        session_status.close()
        conn.close()



from __future__ import annotations

import logging
import os
import tempfile
from typing import TYPE_CHECKING

from sqlalchemy import select, func
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.orm import with_polymorphic

from qcfractal.components.base_dataset_socket import BaseDatasetSocket
from qcfractal.components.dataset_db_models import (
    BaseDatasetORM,
    ContributedValuesORM,
    DatasetAttachmentORM,
    DatasetInternalJobORM,
    DatasetRecordCountORM,
)
from qcfractal.components.dataset_db_views import DatasetDirectRecordsView
from qcfractal.components.dataset_processing import create_view_file
from qcfractal.components.internal_jobs.db_models import InternalJobORM
from qcfractal.db_socket.helpers import (
    get_general,
)
from qcportal.dataset_models import BaseDataset, DatasetAttachmentType
from qcportal.exceptions import MissingDataError, UserReportableError
from qcportal.internal_jobs import InternalJobStatusEnum
from qcportal.metadata_models import InsertMetadata
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.components.internal_jobs.status import JobProgress
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import Dict, Any, Optional, Sequence, Iterable, List, Union
    from typing import Iterable, List, Dict, Any
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from sqlalchemy.orm.session import Session


class DatasetSocket:
    """
    Root socket for all dataset sockets
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

        from qcfractal.components.singlepoint.dataset_socket import SinglepointDatasetSocket
        from qcfractal.components.optimization.dataset_socket import OptimizationDatasetSocket
        from qcfractal.components.torsiondrive.dataset_socket import TorsiondriveDatasetSocket
        from qcfractal.components.gridoptimization.dataset_socket import GridoptimizationDatasetSocket
        from qcfractal.components.manybody.dataset_socket import ManybodyDatasetSocket
        from qcfractal.components.reaction.dataset_socket import ReactionDatasetSocket
        from qcfractal.components.neb.dataset_socket import NEBDatasetSocket

        self.singlepoint = SinglepointDatasetSocket(root_socket)
        self.optimization = OptimizationDatasetSocket(root_socket)
        self.torsiondrive = TorsiondriveDatasetSocket(root_socket)
        self.gridoptimization = GridoptimizationDatasetSocket(root_socket)
        self.manybody = ManybodyDatasetSocket(root_socket)
        self.reaction = ReactionDatasetSocket(root_socket)
        self.neb = NEBDatasetSocket(root_socket)

        self._handler_map: Dict[str, BaseDatasetSocket] = {
            self.singlepoint.dataset_type: self.singlepoint,
            self.optimization.dataset_type: self.optimization,
            self.torsiondrive.dataset_type: self.torsiondrive,
            self.gridoptimization.dataset_type: self.gridoptimization,
            self.manybody.dataset_type: self.manybody,
            self.reaction.dataset_type: self.reaction,
            self.neb.dataset_type: self.neb,
        }

    def get_socket(self, dataset_type: str) -> BaseDatasetSocket:
        """
        Get the socket for a specific kind of dataset type
        """

        handler = self._handler_map.get(dataset_type, None)
        if handler is None:
            raise MissingDataError(f"Cannot find handler for type {dataset_type}")
        return handler

    def get(
        self,
        dataset_id: int,
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Obtain a dataset with the specified ID

        Parameters
        ----------
        dataset_id
            ID of a dataset
        include
            Which fields of the result to return. Default is to return all fields.
        exclude
            Remove these fields from the return. Default is to return all fields.
        missing_ok
           If set to True, None will be returned if the dataset does not exist. Otherwise,
           an exception is raised.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Dataset information as a dictionary
        """

        # If all columns are included, then we can load
        # the data from derived classes as well.
        if (include is None or "*" in include) and not exclude:
            wp = with_polymorphic(BaseDatasetORM, "*")
        else:
            wp = BaseDatasetORM

        with self.root_socket.optional_session(session, True) as session:
            return get_general(session, wp, wp.id, [dataset_id], include, exclude, missing_ok)[0]

    def status(self, dataset_id: int, *, session: Optional[Session] = None) -> Dict[str, Dict[RecordStatusEnum, int]]:
        """
        Compute the status of a dataset

        This function will perform the lookup of the dataset type and the call the dataset-specific socket

        Parameters
        ----------
        dataset_id
            ID of a dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Dictionary with specifications as the keys, and record status/counts as values.
        """

        with self.root_socket.optional_session(session, True) as session:
            ds_type = self.lookup_type(dataset_id, session=session)
            ds_socket = self.get_socket(ds_type)
            return ds_socket.status(dataset_id, session=session)

    def overall_status(self, dataset_id: int, *, session: Optional[Session] = None) -> Dict[RecordStatusEnum, int]:
        """
        Compute the overall status of a dataset

        Similar to the status() function, but is only a dictionary of status and the counts. That is, the status
        is not broken out by specification

        Parameters
        ----------
        dataset_id
            ID of a dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Dictionary with record status as keys and counts as values.
        """

        stat = self.status(dataset_id, session=session)

        overall_stat = {}
        for spec_stats in stat.values():
            for k, v in spec_stats.items():
                overall_stat.setdefault(k, 0)
                overall_stat[k] += v

        return overall_stat

    def lookup_type(self, dataset_id: int, *, session: Optional[Session] = None) -> str:
        """
        Look up the type of dataset given its ID
        """

        stmt = select(BaseDatasetORM.dataset_type)
        stmt = stmt.where(BaseDatasetORM.id == dataset_id)

        with self.root_socket.optional_session(session, True) as session:
            ds_type = session.execute(stmt).scalar_one_or_none()

            if ds_type is None:
                raise MissingDataError(f"Could not find dataset with id {dataset_id}")

            return ds_type

    def lookup_id(
        self, dataset_type: str, dataset_name: str, missing_ok: bool = False, *, session: Optional[Session] = None
    ) -> Optional[int]:
        """
        Look up a dataset ID given its dataset type and name
        """

        stmt = select(BaseDatasetORM.id)
        stmt = stmt.where(BaseDatasetORM.lname == dataset_name.lower())
        stmt = stmt.where(BaseDatasetORM.dataset_type == dataset_type.lower())

        with self.root_socket.optional_session(session, True) as session:
            ds_id = session.execute(stmt).scalar_one_or_none()

            if missing_ok is False and ds_id is None:
                raise MissingDataError(f"Could not find {dataset_type} dataset with name '{dataset_name}'")
            return ds_id

    def list(self, *, session: Optional[Session] = None) -> List[Dict[str, Any]]:
        """
        Get a list of datasets in the database
        """

        with self.root_socket.optional_session(session, True) as session:
            stmt = select(
                BaseDatasetORM.id,
                BaseDatasetORM.dataset_type,
                BaseDatasetORM.name,
                BaseDatasetORM.tagline,
                BaseDatasetORM.description,
                func.coalesce(DatasetRecordCountORM.record_count, 0),
            )
            stmt = stmt.join(DatasetRecordCountORM, DatasetRecordCountORM.dataset_id == BaseDatasetORM.id, isouter=True)
            stmt = stmt.order_by(BaseDatasetORM.id.asc())
            r = session.execute(stmt).all()

            return [
                {
                    "id": x[0],
                    "dataset_type": x[1],
                    "dataset_name": x[2],
                    "tagline": x[3],
                    "description": x[4],
                    "record_count": x[5],
                }
                for x in r
            ]

    def delete(
        self,
        dataset_id: int,
        delete_records: bool,
        *,
        session: Optional[Session] = None,
    ):
        """
        Deletes an entire dataset from the database

        Parameters
        ----------
        dataset_id
            ID of a dataset
        delete_records
            If true, delete all the individual records as well
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        with self.root_socket.optional_session(session) as session:
            ds_type = self.root_socket.datasets.lookup_type(dataset_id, session=session)
            ds_socket = self.root_socket.datasets.get_socket(ds_type)
            return ds_socket.delete_dataset(dataset_id, delete_records, session=session)

    def query_dataset_records(
        self,
        record_id: Iterable[int],
        dataset_type: Optional[Iterable[str]] = None,
        *,
        session: Optional[Session] = None,
    ):
        """
        Query which datasets the specified records belong to

        This returns a dictionary containing the dataset id, type, and name, as well as
        the entry and specification names that the record belongs to
        """

        stmt = select(
            DatasetDirectRecordsView.c.record_id,
            BaseDatasetORM.id,
            BaseDatasetORM.dataset_type,
            BaseDatasetORM.name,
            DatasetDirectRecordsView.c.entry_name,
            DatasetDirectRecordsView.c.specification_name,
        )
        stmt = stmt.join(DatasetDirectRecordsView, BaseDatasetORM.id == DatasetDirectRecordsView.c.dataset_id)
        stmt = stmt.where(DatasetDirectRecordsView.c.record_id.in_(record_id))

        if dataset_type is not None:
            stmt = stmt.where(BaseDatasetORM.dataset_type == dataset_type)

        with self.root_socket.optional_session(session, True) as session:
            ret = session.execute(stmt).all()
            return [
                {
                    "record_id": x[0],
                    "dataset_id": x[1],
                    "dataset_type": x[2],
                    "dataset_name": x[3],
                    "entry_name": x[4],
                    "specification_name": x[5],
                }
                for x in ret
            ]

    def get_contributed_values(self, dataset_id: int, *, session: Optional[Session] = None) -> List[Dict[str, Any]]:
        """
        Get the contributed values for a dataset
        """

        stmt = select(ContributedValuesORM)
        stmt = stmt.where(ContributedValuesORM.dataset_id == dataset_id)

        with self.root_socket.optional_session(session, True) as session:
            cv = session.execute(stmt).scalars().all()
            return [x.model_dict() for x in cv]

    def get_internal_job(
        self,
        dataset_id: int,
        job_id: int,
        *,
        session: Optional[Session] = None,
    ):
        stmt = select(InternalJobORM)
        stmt = stmt.join(DatasetInternalJobORM, DatasetInternalJobORM.internal_job_id == InternalJobORM.id)
        stmt = stmt.where(DatasetInternalJobORM.dataset_id == dataset_id)
        stmt = stmt.where(DatasetInternalJobORM.internal_job_id == job_id)

        with self.root_socket.optional_session(session, True) as session:
            ij_orm = session.execute(stmt).scalar_one_or_none()
            if ij_orm is None:
                raise MissingDataError(f"Job id {job_id} not found in dataset {dataset_id}")
            return ij_orm.model_dict()

    def list_internal_jobs(
        self,
        dataset_id: int,
        status: Optional[Iterable[InternalJobStatusEnum]] = None,
        *,
        session: Optional[Session] = None,
    ):
        stmt = select(InternalJobORM)
        stmt = stmt.join(DatasetInternalJobORM, DatasetInternalJobORM.internal_job_id == InternalJobORM.id)
        stmt = stmt.where(DatasetInternalJobORM.dataset_id == dataset_id)

        if status is not None:
            stmt = stmt.where(InternalJobORM.status.in_(status))

        with self.root_socket.optional_session(session, True) as session:
            ij_orm = session.execute(stmt).scalars().all()
            return [i.model_dict() for i in ij_orm]

    def get_attachments(self, dataset_id: int, *, session: Optional[Session] = None) -> List[Dict[str, Any]]:
        """
        Get the attachments for a dataset
        """

        stmt = select(DatasetAttachmentORM)
        stmt = stmt.where(DatasetAttachmentORM.dataset_id == dataset_id)

        with self.root_socket.optional_session(session, True) as session:
            att = session.execute(stmt).scalars().all()
            return [x.model_dict() for x in att]

    def delete_attachment(self, dataset_id: int, file_id: int, *, session: Optional[Session] = None):
        stmt = select(DatasetAttachmentORM)
        stmt = stmt.where(DatasetAttachmentORM.dataset_id == dataset_id)
        stmt = stmt.where(DatasetAttachmentORM.id == file_id)
        stmt = stmt.with_for_update()

        with self.root_socket.optional_session(session) as session:
            att = session.execute(stmt).scalar_one_or_none()
            if att is None:
                raise MissingDataError(f"Attachment with file id {file_id} not found in dataset {dataset_id}")

            return self.root_socket.external_files.delete(file_id, session=session)

    def attach_file(
        self,
        dataset_id: int,
        attachment_type: DatasetAttachmentType,
        file_path: str,
        file_name: str,
        description: Optional[str],
        provenance: Dict[str, Any],
        *,
        job_progress: Optional[JobProgress] = None,
        session: Optional[Session] = None,
    ) -> int:
        """
        Attach a file to a dataset

        This function uploads the specified file and associates it with the dataset by creating
        a corresponding dataset attachment record. This operation requires S3 storage to be enabled.

        Parameters
        ----------
        dataset_id
            The ID of the dataset to which the file will be attached.
        attachment_type
            The type of attachment that categorizes the file being added.
        file_path
            The local file system path to the file that needs to be uploaded.
        file_name
            The name of the file to be used in the attachment record. This is the filename that is
            recommended to the user by default.
        description
            An optional description of the file
        provenance
            A dictionary containing metadata regarding the origin or history of the file.
        job_progress
            Object used to track progress if this function is being run in a background job
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Raises
        ------
        UserReportableError
            Raised if S3 storage is not enabled
        """

        if not self.root_socket.qcf_config.s3.enabled:
            raise UserReportableError("S3 storage is not enabled. Can not attach file to a dataset")

        self._logger.info(f"Uploading/Attaching dataset-related file: {file_path}")
        with self.root_socket.optional_session(session) as session:
            ef = DatasetAttachmentORM(
                dataset_id=dataset_id,
                attachment_type=attachment_type,
                file_name=file_name,
                description=description,
                provenance=provenance,
            )

            file_id = self.root_socket.external_files.add_file(
                file_path, ef, session=session, job_progress=job_progress
            )

            self._logger.info(f"Dataset attachment {file_path} successfully uploaded to S3. ID is {file_id}")
            return file_id

    def create_view_attachment(
        self,
        dataset_id: int,
        dataset_type: str,
        description: Optional[str],
        provenance: Dict[str, Any],
        status: Optional[Iterable[RecordStatusEnum]] = None,
        include: Optional[Iterable[str]] = None,
        exclude: Optional[Iterable[str]] = None,
        *,
        include_children: bool = True,
        job_progress: Optional[JobProgress] = None,
        session: Optional[Session] = None,
    ):
        """
        Creates a dataset view and attaches it to the dataset

        Uses a temporary directory within the globally-configured `temporary_dir`

        Parameters
        ----------
        dataset_id : int
            ID of the dataset to create the view for
        dataset_type
            Type of dataset the ID is
        description
            Optional string describing the view file
        provenance
            Dictionary with any metadata or other information about the view. Information regarding
            the options used to create the view will be added.
        status
            List of statuses to include. Default is to include records with any status
        include
            List of specific record fields to include in the export. Default is to include most fields
        exclude
            List of specific record fields to exclude from the export. Defaults to excluding none.
        include_children
            Specifies whether child records associated with the main records should also be included (recursively)
            in the view file.
        job_progress
            Object used to track progress if this function is being run in a background job
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        if not self.root_socket.qcf_config.s3.enabled:
            raise UserReportableError("S3 storage is not enabled. Can not not create view")

        # Add the options for the view creation to the provenance
        provenance = provenance | {
            "options": {
                "status": status,
                "include": include,
                "exclude": exclude,
                "include_children": include_children,
            }
        }

        with tempfile.TemporaryDirectory(dir=self.root_socket.qcf_config.temporary_dir) as tmpdir:
            self._logger.info(f"Using temporary directory {tmpdir} for view creation")

            file_name = f"dataset_{dataset_id}_view.sqlite"
            tmp_file_path = os.path.join(tmpdir, file_name)

            create_view_file(
                session,
                self.root_socket,
                dataset_id,
                dataset_type,
                tmp_file_path,
                status=status,
                include=include,
                exclude=exclude,
                include_children=include_children,
                job_progress=job_progress,
            )

            self._logger.info(f"View file created. File size is {os.path.getsize(tmp_file_path)/1048576} MiB.")

            if job_progress is not None:
                job_progress.update_progress(90, "Uploading view file to S3")

            file_id = self.attach_file(
                dataset_id, DatasetAttachmentType.view, tmp_file_path, file_name, description, provenance
            )

            if job_progress is not None:
                job_progress.update_progress(100)

            return file_id

    def add_create_view_attachment_job(
        self,
        dataset_id: int,
        dataset_type: str,
        description: str,
        provenance: Dict[str, Any],
        status: Optional[Iterable[RecordStatusEnum]] = None,
        include: Optional[Iterable[str]] = None,
        exclude: Optional[Iterable[str]] = None,
        *,
        include_children: bool = True,
        session: Optional[Session] = None,
    ) -> int:
        """
        Creates an internal job for creating and attaching a view to a dataset

        See :meth:`create_view_attachment` for a description of the parameters

        Returns
        -------
        :
            ID of the created internal job
        """

        if not self.root_socket.qcf_config.s3.enabled:
            raise UserReportableError("S3 storage is not enabled. Can not not create view")

        with self.root_socket.optional_session(session) as session:
            job_id = self.root_socket.internal_jobs.add(
                f"create_attach_view_ds_{dataset_id}",
                now_at_utc(),
                f"datasets.create_view_attachment",
                {
                    "dataset_id": dataset_id,
                    "dataset_type": dataset_type,
                    "description": description,
                    "provenance": provenance,
                    "status": status,
                    "include": include,
                    "exclude": exclude,
                    "include_children": include_children,
                },
                user_id=None,
                unique_name=True,
                serial_group="ds_create_view",
                session=session,
            )

            stmt = (
                insert(DatasetInternalJobORM)
                .values(dataset_id=dataset_id, internal_job_id=job_id)
                .on_conflict_do_nothing()
            )
            session.execute(stmt)
            return job_id

    def submit(
        self,
        dataset_id: int,
        entry_names: Optional[Iterable[str]],
        specification_names: Optional[Iterable[str]],
        compute_tag: Optional[str],
        compute_priority: Optional[PriorityEnum],
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ):
        """
        Submit computations for a dataset

        This function looks up the dataset socket and then call submit on that socket
        """

        with self.root_socket.optional_session(session) as session:
            ds_type = self.lookup_type(dataset_id)
            ds_socket = self.get_socket(ds_type)

            return ds_socket.submit(
                dataset_id=dataset_id,
                entry_names=entry_names,
                specification_names=specification_names,
                compute_tag=compute_tag,
                compute_priority=compute_priority,
                creator_user=creator_user,
                find_existing=find_existing,
                session=session,
            )

    def add_entry_dicts(
        self, dataset_id: int, entry_dicts: list[dict], *, session: Optional[Session] = None
    ) -> InsertMetadata:
        """
        Add entries to a dataset, where entries are dictionaries

        This function looks up the dataset socket and then casts to the appropriate type,
        calling the add_entries function of that socket
        """

        with self.root_socket.optional_session(session) as session:
            ds_type = self.lookup_type(dataset_id)
            ds_socket = self.get_socket(ds_type)

            # entry types always derive from newentry types
            entry_type = BaseDataset.get_subclass(ds_type)._new_entry_type
            new_entries = [entry_type(**d) for d in entry_dicts]

            return ds_socket.add_entries(
                dataset_id=dataset_id,
                new_entries=new_entries,
                session=session,
            )



import hashlib
import os
import shutil
import subprocess
import time
from typing import List, Optional

import pytest
import yaml

from qcarchivetesting import testconfig_path, migrationdata_path
from qcfractal.config import read_configuration
from qcfractal.port_util import find_open_port

config_file = os.path.join(testconfig_path, "qcf_basic.yaml")

pytestmark = pytest.mark.slow


@pytest.fixture(scope="function")
def unique_db_name(request):
    # Use a hash of the node (which is unique in pytest)
    # Prefix with qcf_ to make sure it is a valid dbname
    yield f"qcf_{hashlib.sha1(request.node.nodeid.encode()).hexdigest()}"


@pytest.fixture(scope="function")
def tmp_config(tmp_path, unique_db_name):
    full_config_path = os.path.join(tmp_path, os.path.basename(config_file))

    # inject the unique database name (unique per test)
    with open(config_file, "r") as f:
        config = yaml.safe_load(f)

    config["database"]["database_name"] = unique_db_name
    config["database"]["port"] = find_open_port()
    config["api"]["port"] = find_open_port()

    with open(full_config_path, "w") as f:
        yaml.dump(config, f)

    yield full_config_path


# Test with both owning the database and not owning it
@pytest.fixture(scope="function", params=[True, False])
def cli_runner_core(postgres_server, tmp_config, request, unique_db_name):

    pg_harness = None
    own_db = request.param

    if not own_db:
        # Unique database for each test
        # This is needed when running tests in parallel
        pg_harness = postgres_server.get_new_harness(unique_db_name)

        # Don't actually use the database - we just want a placeholder
        pg_harness.delete_database()

        config = read_configuration([tmp_config])
        config.database = pg_harness.config.model_copy(deep=True)
        config.database.own = False

        with open(tmp_config, "w") as f:
            yaml.dump(config.model_dump(), f)

    # Use a functor so we can get own_db (and maybe other info in the future)
    class run_qcfractal_cli:
        def __init__(self):
            self.own_db = own_db
            self.db_name = unique_db_name
            self.config_path = tmp_config

        def __call__(
            self,
            cmd_args: List[str],
            *,
            fail_expected=False,
            timeout_expected=False,
            stdin: Optional[str] = None,
            timeout: Optional[int] = None,
        ):
            full_cmd = ["qcfractal-server", "-v", "--config", self.config_path] + cmd_args
            try:
                ret = subprocess.run(full_cmd, capture_output=True, text=True, input=stdin, timeout=timeout)
            except subprocess.TimeoutExpired as err:
                if timeout_expected:
                    return err.output
                else:
                    raise

            if ret.returncode == 0 and not fail_expected:
                return ret.stdout + ret.stderr
            if ret.returncode != 0 and fail_expected:
                return ret.stdout + ret.stderr
            if ret.returncode == 0 and fail_expected:
                raise RuntimeError("Failure expected, but process succeeded")
            if ret.returncode != 0 and not fail_expected:
                raise RuntimeError("Error running process:\n" + ret.stdout + "\n" + ret.stderr)

    try:
        yield run_qcfractal_cli()
    finally:
        if pg_harness is not None:
            pg_harness.delete_database()


@pytest.fixture(scope="function")
def cli_runner(cli_runner_core):
    # Creates the postgres instance (if needed) and tables
    cli_runner_core(["init-db"])

    yield cli_runner_core


def test_cli_help():
    subprocess.check_output(["qcfractal-server", "--help"])


def test_cli_info_config(cli_runner):
    output = cli_runner(["info", "server"])
    assert "THISISASECRETKEY" in output
    assert "QCFractal Test Server" in output


def test_cli_info_alembic(cli_runner_core):
    output = cli_runner_core(["info", "alembic"])
    assert "alembic -c" in output


@pytest.mark.parametrize("full", [True, False])
def test_cli_init_config(cli_runner_core, full: bool):
    # Remove the old config
    os.remove(cli_runner_core.config_path)

    cmd = ["init-config"]
    if full:
        cmd.append("--full")

    output = cli_runner_core(cmd)
    assert "Creating initial QCFractal configuration" in output

    # Replace the port in the config file (to make this safe for parallel tests)
    with open(cli_runner_core.config_path, "r") as f:
        config = yaml.safe_load(f)
    config["database"]["port"] = find_open_port()
    with open(cli_runner_core.config_path, "w") as f:
        yaml.dump(config, f)

    # Initialize the db. Needed for the config command
    cli_runner_core(["init-db"])

    # Tests reading the config
    output = cli_runner_core(["info", "server"])
    assert "name: QCFractal Server" in output

    with open(cli_runner_core.config_path, "r") as f:
        config_text = f.read()

    if full:
        assert "hide_internal_errors" in config_text
    else:
        assert "hide_internal_errors" not in config_text


def test_cli_init(cli_runner_core):
    output = cli_runner_core(["init-db"])

    assert "Initializing QCFractal database from configuration" in output

    if cli_runner_core.own_db is True:
        assert "Success. You can now start the database server using" in output
        assert "pg_ctl: server is running" in output
        assert "PostgreSQL successfully stopped" in output
    else:
        assert "Success. You can now start the database server using" not in output
        assert "pg_ctl: server is running" not in output
        assert "PostgreSQL successfully stopped" not in output


def test_cli_user_add_info_list(cli_runner):
    output = cli_runner(["user", "add", "testuser", "--role", "admin", "--fullname", "A. Test User"])
    assert "Created user testuser" in output

    output = cli_runner(["user", "info", "testuser"])
    assert "testuser" in output
    assert "admin" in output
    assert "A. Test User" in output

    output = cli_runner(["user", "list"])
    assert "testuser" in output
    assert "admin" in output
    assert "A. Test User" in output


def test_cli_user_delete_prompt(cli_runner):
    output = cli_runner(["user", "add", "testuser", "--role", "admin", "--fullname", "A. Test User"])
    assert "Created user testuser" in output

    output = cli_runner(["user", "delete", "testuser"], stdin="N")
    assert "ABORTING" in output

    output = cli_runner(["user", "list"])
    assert "testuser" in output

    output = cli_runner(["user", "delete", "testuser"], stdin="Y")
    assert "User testuser deleted"
    assert "Deleted!" in output

    output = cli_runner(["user", "list"])
    assert "testuser" not in output


def test_cli_user_delete_noprompt(cli_runner):
    output = cli_runner(["user", "add", "testuser", "--role", "admin", "--fullname", "A. Test User"])
    assert "Created user testuser" in output

    output = cli_runner(["user", "delete", "testuser", "--no-prompt"])
    assert "User testuser deleted"
    assert "Deleted!" in output


def test_cli_user_modify_basic(cli_runner):
    output = cli_runner(["user", "add", "testuser", "--role", "admin", "--fullname", "A. Test User"])
    assert "Created user testuser" in output

    output = cli_runner(
        [
            "user",
            "modify",
            "testuser",
            "--fullname",
            "A. New Name",
            "--organization",
            "An Organization",
            "--email",
            "testuser@example.com",
            "--role",
            "submit",
        ]
    )

    assert "User testuser modified" in output

    output = cli_runner(["user", "info", "testuser"])
    assert "A. New Name" in output
    assert "An Organization" in output
    assert "testuser@example.com" in output
    assert "submit" in output
    assert "admin" not in output


def test_cli_user_enable_disable(cli_runner):
    output = cli_runner(["user", "add", "testuser", "--role", "admin", "--fullname", "A. Test User"])
    assert "Created user testuser" in output

    output = cli_runner(["user", "modify", "testuser", "--disable"])
    assert "User testuser modified" in output

    output = cli_runner(["user", "info", "testuser"])
    assert "enabled: False" in output

    output = cli_runner(["user", "modify", "testuser", "--enable"])
    assert "User testuser modified" in output

    output = cli_runner(["user", "info", "testuser"])
    assert "enabled: True" in output


def test_cli_user_password(cli_runner):
    output = cli_runner(["user", "add", "testuser", "--role", "admin", "--fullname", "A. Test User"])
    assert "Created user testuser" in output

    output = cli_runner(["user", "modify", "testuser", "--reset-password"])
    assert "User testuser modified" in output
    assert "New password is below"

    output = cli_runner(["user", "modify", "testuser", "--password", "new_password_1234"])
    assert "User testuser modified" in output
    assert "Password for testuser modified"


def test_cli_restore_noinit(cli_runner_core):
    # Restore where the db does not exist and has not been initialized
    migdata_path = os.path.join(migrationdata_path, "empty_v0.50.sql_dump")
    output = cli_runner_core(["restore", migdata_path])
    assert "Restore complete!" in output

    if cli_runner_core.own_db:
        assert "Postgresql instance successfully initialized and started" in output
    else:
        assert "Postgresql instance successfully initialized and started" not in output


def test_cli_backup_restore(cli_runner_core, tmp_path):
    migdata_path = os.path.join(migrationdata_path, "empty_v0.50.sql_dump")

    cli_runner_core(["restore", migdata_path])
    cli_runner_core(["upgrade-db"])

    backup_dir = tmp_path / "db_bak"
    backup_dir.mkdir()
    backup_path = str(backup_dir / "backup_file.sqlback")

    output = cli_runner_core(["backup", backup_path])
    assert os.path.isfile(backup_path)
    assert "Backup complete!" in output

    db_name = cli_runner_core.db_name
    output = cli_runner_core(["restore", backup_path], stdin=f"REMOVEALLDATA {db_name}")
    assert "Restore complete!" in output


def test_cli_restore_existing(cli_runner):
    # Restore where the db already exists
    migdata_path = os.path.join(migrationdata_path, "empty_v0.50.sql_dump")

    db_name = cli_runner.db_name
    output = cli_runner(["restore", migdata_path], stdin=f"REMOVEALLDATA {db_name}")
    assert "Restore complete!" in output

    migdata_path = os.path.join(migrationdata_path, "empty_v0.50.sql_dump")
    output = cli_runner(["restore", migdata_path], stdin="ASD")
    assert "does not match. Aborting" in output


def test_cli_upgrade(cli_runner_core):
    migdata_path = os.path.join(migrationdata_path, "empty_v0.50.sql_dump")

    db_name = cli_runner_core.db_name
    output = cli_runner_core(["restore", migdata_path], stdin=f"REMOVEALLDATA {db_name}")
    assert "Restore complete!" in output

    output = cli_runner_core(["upgrade-db"])

    # One of the migrations that should be there
    assert "Running upgrade d1ee87a66b71 -> 13cb230def11" in output

    if cli_runner_core.own_db is True:
        assert "PostgreSQL successfully stopped" in output


def test_cli_upgrade_noinit(cli_runner_core):
    output = cli_runner_core(["upgrade-db"], fail_expected=True)
    if cli_runner_core.own_db:
        assert "has not been initialized" in output
    else:
        assert "does not exist for upgrading" in output


def test_cli_start(cli_runner):
    full_cmd = ["qcfractal-server", "--config", cli_runner.config_path, "start"]

    # Manually start then kill
    proc = subprocess.Popen(full_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
    time.sleep(20)
    proc.terminate()
    proc.wait(10)

    assert proc.returncode == 0
    output = proc.stdout.read()

    assert "waitress: Serving on" in output


def test_cli_start_options(cli_runner, tmp_path):

    log_dir = tmp_path / "logs"
    log_dir.mkdir()

    log_path = str(log_dir / "qca_test.logfile")

    port = find_open_port()
    full_cmd = [
        "qcfractal-server",
        "--config",
        cli_runner.config_path,
        "start",
        "--host",
        "0.0.0.0",
        "--port",
        str(port),
        "--logfile",
        log_path,
    ]

    # Manually start then kill
    proc = subprocess.Popen(full_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
    time.sleep(20)
    proc.terminate()
    proc.wait(10)

    assert proc.returncode == 0

    with open(log_path, "r") as f:
        log_output = f.read()

    assert f"waitress: Serving on http://0.0.0.0:{port}" in log_output


def test_cli_start_outdated(cli_runner_core):
    migdata_path = os.path.join(migrationdata_path, "empty_v0.50.sql_dump")

    db_name = cli_runner_core.db_name
    output = cli_runner_core(["restore", migdata_path], stdin=f"REMOVEALLDATA {db_name}")
    assert "Restore complete!" in output

    output = cli_runner_core(["start"], timeout=20, fail_expected=True)
    assert "Database needs migration" in output
